// ==============================================================
//
// titan_dragonfly.cpp
// Control module for titan_dragonfly vessel class
//
// ==============================================================

#define STRICT 1
#define ORBITER_MODULE

#include "orbitersdk.h"
#include "OrbiterSoundSDK40.h"
#include <cstdio>
#include <cmath>
#include <string>
#include <stdio.h>

// ==============================================================
// Some vessel parameters
// ==============================================================
//const DWORD COMM_SET_TGT = 2306;

//#define STRT	1
#define MOTR	1

const double EMPTY_MASS = 449;
const double FUEL_MASS = 1;
const double MAIN_TH = 1500;
const double MAIN_ISP = 6000000;
const double YAW_TH = 150;
const double RTR_OFF = 0.75;


double stiffness = 1e4;
double damping = 1e3;
static const DWORD ntdvtx_skis = 7;
static TOUCHDOWNVTX tdvtx_skis[ntdvtx_skis] = {
	{_V( 0.000,-0.834, 0.763), stiffness, damping, 2.9, 2.9},
	{_V(-0.735,-0.834,-0.763), stiffness, damping, 2.9, 2.9},
	{_V( 0.735,-0.834,-0.763), stiffness, damping, 2.9, 2.9},
	{_V(-0.735,-0.834, 0.763), stiffness, damping, 2.9},
	{_V( 0.735,-0.834, 0.763), stiffness, damping, 2.9},
	{_V( 0.000,-0.050, 1.655), stiffness, damping, 2.9},
	{_V( 0.000, 0.200,-1.400), stiffness, damping, 2.9}
};


void VLiftCoeff (double aoa, double M, double Re, double *cl, double *cm, double *cd)
{
	static const double step = RAD*22.5;
	static const double istep = 1.0/step;
	static const int nabsc = 17;
	static const double CL[nabsc] = {0, 0.1, 0.2, 0.1, 0, 0.1, 0.2, 0.1, 0, -0.1, -0.2, -0.1, 0, -0.1, -0.2, -0.1, 0};


	aoa += PI;
	int idx = max (0, min (15, (int)(aoa*istep)));
	double d = aoa*istep - idx;
	*cl = 2*(CL[idx] + (CL[idx+1]-CL[idx])*d);
	*cm = 0.0;
	*cd = 2 + oapiGetInducedDrag (*cl, 0.16, 0.2);
	*cl = 0;
	
}

// 2. horizontal lift component (vertical stabiliser and body)

void HLiftCoeff (double beta, double M, double Re, double *cl, double *cm, double *cd)
{
	static const double step = RAD*22.5;
	static const double istep = 1.0/step;
	static const int nabsc = 17;
	static const double CL[nabsc] = {0, 0.1, 0.2, 0.1, 0.0, 0.1, 0.2, 0.1, 0, -0.1, -0.2, -0.1, 0,-0.1,-0.2,-0.1, 0};
	beta += PI;
	int idx = max (0, min (15, (int)(beta*istep)));
	double d = beta*istep - idx;
	*cl = -8*(CL[idx] + (CL[idx+1]-CL[idx])*d);
	*cm = 0.0;
	*cd = 0.025*(0.11 + oapiGetInducedDrag (*cl, 0.16, 0.2));
	//sprintf(oapiDebugString(), "beta %0.3f", beta*DEG);

}



// ==============================================================
// Shuttle-PB class interface
// ==============================================================

class titan_dragonfly: public VESSEL3 {
public:
	titan_dragonfly (OBJHANDLE hVessel, int flightmodel);
	~titan_dragonfly ();
	void clbkSetClassCaps (FILEHANDLE cfg);
	void clbkPreStep(double SimT, double SimDT, double MJD);
	void clbkPostStep (double simt, double simdt, double mjd);
	void clbkSaveState (FILEHANDLE scn);
	void clbkLoadStateEx (FILEHANDLE scn, void *status);
	int clbkConsumeBufferedKey(DWORD key, bool down, char *kstate);
	int clbkConsumeDirectKey(char *kstate);
	bool clbkDrawHUD (int mode, const HUDPAINTSPEC *hps, oapi::Sketchpad *skp);
	void clbkPostCreation();



private:

	void SetMainTh(void);
	void DoHoldPos(void);
	void DoHoldAlt(void);
	void DoBoom(void);
	void DefineAnimations (void);
	void DoHGA(void);
	int MyID;
	void DoBatteryCharge(void);
	void DoAutoLand(void);
	//char tgt_name[256];

	THRUSTER_HANDLE th_main[1], th_yaw[2], th_boom[2];
	THGROUP_HANDLE htgrpmain, th_group[1];
	PROPELLANT_HANDLE hProp;
	ATTACHMENTHANDLE hAtt; 
	SpotLight *spotlight1;

	ANIMATIONCOMPONENT_HANDLE Section0;
	MGROUP_TRANSFORM *rtr1, *rtr2, *rtr3, *rtr4, *rtr5, *rtr6, *rtr7, *rtr8;
	MGROUP_TRANSFORM *blur1, *blur2, *blur3, *blur4, *blur5, *blur6, *blur7, *blur8;
	MGROUP_TRANSFORM *ski1, *ski2, *ant1, *ant2;

	int eng_status, auto_hvr_status, auto_alt_status, ski_status, ant_status;
	double step_time, pitch_level, bank_level, yaw_level;// rotor_trim, elev_trim;
	double drot, dthrottle, boom_sw;
	int pitch_mode, bank_mode, yaw_mode, throttle_mode, autoland_status;

	UINT anim_rtr, anim_blur, anim_ski, anim_ant_alt, anim_ant_az;
	double rtr_proc, blur_proc, ski_proc, ant_alt_proc, ant_az_proc;

	//double ground_hdg, ground_spd;

	double js_pu, js_pd, js_bl, js_br, js_yl, js_yr;
	double js_pitch_level, js_bank_level, js_yaw_level;


};

titan_dragonfly::titan_dragonfly (OBJHANDLE hVessel, int flightmodel)
: VESSEL3 (hVessel, flightmodel)
{
    

}

titan_dragonfly::~titan_dragonfly ()
{
	delete rtr1;
	delete rtr2;
	delete rtr3;
	delete rtr4;
	delete rtr5;
	delete rtr6;
	delete rtr7;
	delete rtr8;

	
	delete blur1;
	delete blur2;
	delete blur3;
	delete blur4;
	delete blur5;
	delete blur6;
	delete blur7;
	delete blur8;

	delete ski1;
	delete ski2;

	delete ant1;
	delete ant2;
	
}

void titan_dragonfly::DefineAnimations (void)
{
	
// props
	anim_rtr = CreateAnimation (0);
	static UINT RtrGrp1[1] = {0};
	rtr1 = new MGROUP_ROTATE (1, RtrGrp1, 1, _V( 1.021,0.314, 0.643), _V(-0.0872,0.9962,0), (float)(360*RAD));
	AddAnimationComponent (anim_rtr, 0, 1, rtr1);

	static UINT RtrGrp2[1] = {3};
	rtr2 = new MGROUP_ROTATE (1, RtrGrp2, 1, _V( 1.021,0.314,-0.643), _V(-0.0872,0.9962,0), (float)(-360*RAD));
	AddAnimationComponent (anim_rtr, 0, 1, rtr2);

	static UINT RtrGrp3[1] = {7};
	rtr3 = new MGROUP_ROTATE (1, RtrGrp3, 1, _V(-1.021,0.314,-0.643), _V(0.0872,0.9962,0), (float)(360*RAD));
	AddAnimationComponent (anim_rtr, 0, 1, rtr3);

	static UINT RtrGrp4[1] = {5};
	rtr4 = new MGROUP_ROTATE (1, RtrGrp4, 1, _V(-1.021,0.314, 0.643), _V(0.0872,0.9962,0), (float)(-360*RAD));
	AddAnimationComponent (anim_rtr, 0, 1, rtr4);

	static UINT RtrGrp5[1] = {1};
	rtr5 = new MGROUP_ROTATE (1, RtrGrp5, 1, _V( 1.021,0.314, 0.643), _V(-0.0872,0.9962,0), (float)(-360*RAD));
	AddAnimationComponent (anim_rtr, 0, 1, rtr5);

	static UINT RtrGrp6[1] = {2};
	rtr6 = new MGROUP_ROTATE (1, RtrGrp6, 1, _V( 1.021,0.314,-0.643), _V(-0.0872,0.9962,0), (float)( 360*RAD));
	AddAnimationComponent (anim_rtr, 0, 1, rtr6);

	static UINT RtrGrp7[1] = {6};
	rtr7 = new MGROUP_ROTATE (1, RtrGrp7, 1, _V(-1.021,0.314,-0.643), _V(0.0872,0.9962,0), (float)(-360*RAD));
	AddAnimationComponent (anim_rtr, 0, 1, rtr7);

	static UINT RtrGrp8[1] = {4};
	rtr8 = new MGROUP_ROTATE (1, RtrGrp8, 1, _V(-1.021,0.314, 0.643), _V(0.0872,0.9962,0), (float)( 360*RAD));
	AddAnimationComponent (anim_rtr, 0, 1, rtr8);

	////////////////// blur
	anim_blur = CreateAnimation (0);
	static UINT BlurGrp1[1] = {3};
	blur1 = new MGROUP_ROTATE (2, BlurGrp1, 1, _V( 1.021,0.314, 0.643), _V(-0.0872,0.9962,0), (float)(360*RAD));
	AddAnimationComponent (anim_blur, 0, 1, blur1);

	static UINT BlurGrp2[1] = {1};
	blur2 = new MGROUP_ROTATE (2, BlurGrp2, 1, _V( 1.021,0.314,-0.643), _V(-0.0872,0.9962,0), (float)(-360*RAD));
	AddAnimationComponent (anim_blur, 0, 1, blur2);

	static UINT BlurGrp3[1] = {4};
	blur3 = new MGROUP_ROTATE (2, BlurGrp3, 1, _V(-1.021,0.314,-0.643), _V(0.0872,0.9962,0), (float)(360*RAD));
	AddAnimationComponent (anim_blur, 0, 1, blur3);

	static UINT BlurGrp4[1] = {7};
	blur4 = new MGROUP_ROTATE (2, BlurGrp4, 1, _V(-1.021,0.314, 0.643), _V(0.0872,0.9962,0), (float)(-360*RAD));
	AddAnimationComponent (anim_blur, 0, 1, blur4);

	static UINT BlurGrp5[1] = {2};
	blur5 = new MGROUP_ROTATE (2, BlurGrp5, 1, _V( 1.021,0.314, 0.643), _V(-0.0872,0.9962,0), (float)(-360*RAD));
	AddAnimationComponent (anim_blur, 0, 1, blur5);

	static UINT BlurGrp6[1] = {0};
	blur6 = new MGROUP_ROTATE (2, BlurGrp6, 1, _V( 1.021,0.314,-0.643), _V(-0.0872,0.9962,0), (float)( 360*RAD));
	AddAnimationComponent (anim_blur, 0, 1, blur6);

	static UINT BlurGrp7[1] = {5};
	blur7 = new MGROUP_ROTATE (2, BlurGrp7, 1, _V(-1.021,0.314,-0.643), _V(0.0872,0.9962,0), (float)(-360*RAD));
	AddAnimationComponent (anim_blur, 0, 1, blur7);

	static UINT BlurGrp8[1] = {6};
	blur8 = new MGROUP_ROTATE (2, BlurGrp8, 1, _V(-1.021,0.314, 0.643), _V(0.0872,0.9962,0), (float)( 360*RAD));
	AddAnimationComponent (anim_blur, 0, 1, blur8);

	//////////////////////////////////////////// skis

	anim_ski = CreateAnimation (0);
	static UINT SkiGrp1[4] = {19,20,25,26};
	ski1 = new MGROUP_ROTATE (0, SkiGrp1, 4, _V(-0.557, -0.517, 0), _V(0,0,1), (float)(135*RAD));
	AddAnimationComponent (anim_ski, 0, 1, ski1);

	static UINT SkiGrp2[4] = {11,12,21,23};
	ski2 = new MGROUP_ROTATE (0, SkiGrp2, 4, _V( 0.557, -0.517, 0), _V(0,0,1), (float)(-135*RAD));
	AddAnimationComponent (anim_ski, 0, 1, ski2);

	///////////////////////// hga //////////////////////////////
	
	anim_ant_az = CreateAnimation (0);
	static UINT AntGrp1[2] = {9,10};
	ant1 = new MGROUP_ROTATE (0, AntGrp1, 2, _V(0, 0, 0.591), _V(0,1,0), (float)(360*RAD));
	Section0 = AddAnimationComponent (anim_ant_az, 0, 1, ant1);

	anim_ant_alt = CreateAnimation (0);
	static UINT AntGrp2[6] = {15,16,17,18,31,32};
	ant2 = new MGROUP_ROTATE (0, AntGrp2, 6, _V(0, 0.442, 0.591), _V(1,0,0), (float)(180*RAD));
	AddAnimationComponent (anim_ant_alt, 0, 1, ant2, Section0);

}


// ==============================================================
// Overloaded callback functions
// ==============================================================

// --------------------------------------------------------------
// Set the capabilities of the vessel class
// --------------------------------------------------------------
void titan_dragonfly::clbkSetClassCaps (FILEHANDLE cfg)
{
	eng_status = 0;
	step_time = 0;
	pitch_level = 0;
	bank_level = 0;
	yaw_level = 0;
	rtr_proc = 0;
	blur_proc = 0;
	pitch_mode = 0;
	bank_mode = 0;
	yaw_mode = 0;
	throttle_mode = 0;
	dthrottle = 0;
	autoland_status = 0;

	js_pu = 0; // joystick input
	js_pd = 0;
	js_bl = 0;
	js_br = 0;
	js_yl = 0;
	js_yr = 0;
	js_pitch_level = 0;
	js_bank_level = 0;
	js_yaw_level = 0;

	drot = 0;

	auto_hvr_status = 0;
	auto_alt_status = 0;

	ski_status = 2;
	ski_proc = 1;

	ant_status = 0;
	ant_alt_proc = 0;
	ant_az_proc = 0;

	boom_sw = 0;

	DefineAnimations ();
	// physical vessel parameters
	SetSize (2);
	SetEmptyMass (EMPTY_MASS);
	SetPMI(_V(1,1,1));
	SetCrossSections (_V(2,3,1));
	SetCW(1,1,1,1);
	SetRotDrag (_V(0.07,0.07,0.07));
	SetTouchdownPoints ( tdvtx_skis, ntdvtx_skis);

	// camera parameters
	SetCameraOffset (_V( 0.25, 0.225, 1.625));
	SetMeshVisibilityMode(AddMesh("titan_dragonfly/titan_dragonfly"),MESHVIS_ALWAYS); // 0
	SetMeshVisibilityMode(AddMesh("titan_dragonfly/titan_dragonfly_prop"),MESHVIS_ALWAYS); // 1
	SetMeshVisibilityMode(AddMesh("titan_dragonfly/titan_dragonfly_blur"),MESHVIS_NEVER); // 2
	

	// ***************** thruster definitions *******************

	hProp = CreatePropellantResource (FUEL_MASS,FUEL_MASS);
	
	th_main[0] = CreateThruster (_V( 0, RTR_OFF, 0), _V( 0,1,0), MAIN_TH, hProp, MAIN_ISP);

	th_group[0] = th_main[0];
	htgrpmain = CreateThrusterGroup (th_group, 1, THGROUP_MAIN);

	th_yaw[0] = CreateThruster (_V( 0, 0, -1), _V( 1,0,0), YAW_TH, hProp, MAIN_ISP);
	th_yaw[1] = CreateThruster (_V( 0, 0, -1), _V(-1,0,0), YAW_TH, hProp, MAIN_ISP);

	th_group[0] = th_yaw[0];
	CreateThrusterGroup (th_group, 1, THGROUP_ATT_YAWLEFT);

	th_group[0] = th_yaw[1];
	CreateThrusterGroup (th_group, 1, THGROUP_ATT_YAWRIGHT);

	///////////////////////////////////////////////////////
	/////AirFoils

	CreateAirfoil2 (LIFT_VERTICAL, _V(0,-0.0,0), VLiftCoeff, 0.125, 0.125, 0.5);
	CreateAirfoil2 (LIFT_HORIZONTAL, _V(0,0,-2.0), HLiftCoeff, 0.125, 0.125, 1);

	//CreateControlSurface3(AIRCTRL_ELEVATORTRIM, 0.025, 0.025, _V(0,0,-1), AIRCTRL_AXIS_XPOS);

	/////////////// attach pnt
	hAtt = CreateAttachment (true, _V(0,0, 0), _V(0,1,0), _V(0,0,1), "TD");

	///////////////////////////// boom p streams

	th_boom[0] = CreateThruster (_V( 0, 0, 0), _V( 0,0,1), 0.001, hProp, 2, 1);
	th_boom[1] = CreateThruster (_V( 0, 0, 0), _V( 0,0,1), 0.001, hProp, 2, 1);

	SURFHANDLE boom1 = oapiRegisterParticleTexture ("Contrail4");
	SURFHANDLE boom2 = oapiRegisterParticleTexture ("Contrail3");

	PARTICLESTREAMSPEC boom_pst1 = {
		0, 8, 100, 150, 2, 2, 10, 4, PARTICLESTREAMSPEC::DIFFUSE,
		PARTICLESTREAMSPEC::LVL_SQRT, 0, 1,
		PARTICLESTREAMSPEC::ATM_PLOG, 1e-7, 1,
		boom1
	};

	PARTICLESTREAMSPEC boom_pst2 = {
		0, 8, 80, 100, 2, 1, 15, 2, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_SQRT, 0, 1,
		PARTICLESTREAMSPEC::ATM_PLOG, 1e-7, 1,
		boom2
	};

	AddExhaustStream(th_boom[0],_V(0,0,0),&boom_pst1);
	AddExhaustStream(th_boom[1],_V(0,0,0),&boom_pst2);
}



// ==============================================================
// API callback interface
// ==============================================================

// --------------------------------------------------------------
// Vessel initialisation
// --------------------------------------------------------------
DLLCLBK VESSEL *ovcInit (OBJHANDLE hvessel, int flightmodel)
{
	return new titan_dragonfly (hvessel, flightmodel);
}

// --------------------------------------------------------------
// Vessel cleanup
// --------------------------------------------------------------
DLLCLBK void ovcExit (VESSEL *vessel)
{
	if (vessel) delete (titan_dragonfly*)vessel;
}




void titan_dragonfly::clbkSaveState (FILEHANDLE scn)
{
	char cbuf[256];
	//elev_trim = GetControlSurfaceLevel(AIRCTRL_ELEVATORTRIM);

	// default vessel parameters
	VESSEL3::clbkSaveState(scn);

	sprintf (cbuf, "%d %0.4f %0.4f %0.4f", eng_status, rtr_proc, blur_proc, drot);
	oapiWriteScenario_string (scn, "ROTR", cbuf);

	sprintf (cbuf, "%d %d %d %0.4f", auto_hvr_status, auto_alt_status, autoland_status, boom_sw);
	oapiWriteScenario_string (scn, "AUTO", cbuf);

	sprintf (cbuf, "%d %0.4f", ski_status, ski_proc);
	oapiWriteScenario_string (scn, "SKIS", cbuf);

	sprintf (cbuf, "%d %0.4f %0.4f", ant_status, ant_alt_proc, ant_az_proc);
	oapiWriteScenario_string (scn, "HGA", cbuf);

}


void titan_dragonfly::clbkLoadStateEx (FILEHANDLE scn, void *vs)
{
    char *line;
	while (oapiReadScenario_nextline (scn, line)) {
		if (!strnicmp (line, "ROTR", 4)) 
		{
			sscanf (line+4, "%d%lf%lf%lf", &eng_status, &rtr_proc, &blur_proc, &drot);
		} 
		else if (!strnicmp (line, "AUTO", 4)) 
		{
			sscanf (line+4, "%d%d%d%lf", &auto_hvr_status, &auto_alt_status, &autoland_status, &boom_sw);
		}
		else if (!strnicmp (line, "SKIS", 4)) 
		{
			sscanf (line+4, "%d%lf", &ski_status, &ski_proc);
		}
		else if (!strnicmp (line, "HGA", 3)) 
		{
			sscanf (line+3, "%d%lf%lf", &ant_status, &ant_alt_proc, &ant_az_proc);
		}
		else 
		{
            ParseScenarioLineEx (line, vs);
			// unrecognised option - pass to Orbiter's generic parser
        }
   }
	
	//SetControlSurfaceLevel(AIRCTRL_ELEVATORTRIM, elev_trim, true);
	SetAnimation(anim_rtr, rtr_proc);
	SetAnimation(anim_blur, blur_proc);
	SetAnimation(anim_ski, ski_proc);
	SetAnimation(anim_ant_alt, ant_alt_proc);
	SetAnimation(anim_ant_az, ant_az_proc);
	
}



int titan_dragonfly::clbkConsumeBufferedKey(DWORD key, bool down, char *kstate)
{

  if (!down) return 0; // only process keydown events

 if (KEYMOD_SHIFT (kstate)) 
  {	 
	switch (key)
	{
		

	} 
  }


 else if (KEYMOD_CONTROL (kstate)) 
  { 
	switch (key)
	{
		
	} 
  }

else { // unmodified keys
    switch (key) 
		{
		case OAPI_KEY_K: // engine start / stop
		//if(boom_sw == 0)
		//{
		//if (eng_status == 0) eng_status = 1; // wind up
		//else if (eng_status == 1) eng_status = 3; // wind down
		//else if (eng_status == 2)
		//		{
		//		eng_status = 3;
		//		SetMeshVisibilityMode(1,MESHVIS_ALWAYS); // 1
		//		SetMeshVisibilityMode(2,MESHVIS_NEVER); // 2
		//		}
		//else if (eng_status == 3) eng_status = 1;
		//}
		return 1;

		case OAPI_KEY_G: // skis
			if(ski_status == 2)ski_status = 1;
			//else if(ski_status == 0)ski_status = 2;
		return 1;

		case OAPI_KEY_M: // ant
			if(ant_status == 0)ant_status = 1;
			else ant_status = 3;
		return 1;

		case OAPI_KEY_N: // autoland
			if(autoland_status == 0)
			{	
			auto_hvr_status = 0;
			auto_alt_status = 0;
			autoland_status = 1;
			SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT,0);
			SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT,0);
			}
			else 
			{
			autoland_status = 0;
			SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT,0);
			SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT,0);
			}
		return 1;

		case OAPI_KEY_V: // auto hover
		if (auto_hvr_status == 0) 
		{	
			auto_hvr_status = 1;
			auto_alt_status = 0;
			autoland_status = 0;
			SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT,0);
			SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT,0);
		}
		else if (auto_hvr_status == 1)
		{
			SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT,0);
			SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT,0);
			auto_hvr_status = 0;
		}
		return 1;

		case OAPI_KEY_B: // auto hold alt
		if (auto_alt_status == 0) 
		{
			auto_alt_status = 1;
			auto_hvr_status = 0;
			autoland_status = 0;
			SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT,0);
			SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT,0);
		}
		else if (auto_alt_status == 1)
		{
			auto_alt_status = 0;
			SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT,0);
			SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT,0);
		}
		return 1;



			
			case OAPI_KEY_NUMPAD5: 
			RESETKEY (kstate, OAPI_KEY_NUMPAD5);
			return 1;
		}
  }
  return 0;
}


int titan_dragonfly::clbkConsumeDirectKey(char *kstate)
{
	if (KEYMOD_SHIFT (kstate))
	{
		//if ((auto_alt_status + auto_hvr_status) == 0)
		//{
		if (KEYDOWN (kstate, OAPI_KEY_ADD)) 
			{
			RESETKEY (kstate, OAPI_KEY_ADD);
			throttle_mode = 1;
			}

		if (KEYDOWN (kstate, OAPI_KEY_SUBTRACT)) 
			{
			RESETKEY (kstate, OAPI_KEY_SUBTRACT);
			throttle_mode = -1;
			}
		//}

	}
	else if (KEYMOD_CONTROL (kstate))
	{
		if (eng_status == 0)
		{
			//RESETKEY (kstate, OAPI_KEY_ADD);
			RESETKEY (kstate, OAPI_KEY_NUMPAD1);
			RESETKEY (kstate, OAPI_KEY_NUMPAD3);
		}

		if ((auto_alt_status + auto_hvr_status) > 0)
		{
			if (KEYDOWN (kstate, OAPI_KEY_ADD)) 
			{
			RESETKEY (kstate, OAPI_KEY_ADD);
			throttle_mode = 1;
			}

		if (KEYDOWN (kstate, OAPI_KEY_SUBTRACT)) 
			{
			RESETKEY (kstate, OAPI_KEY_SUBTRACT);
			throttle_mode = -1;
			}
		}

	}
	else
	{
		if (eng_status == 0)
		{
			//RESETKEY (kstate, OAPI_KEY_ADD);
			RESETKEY (kstate, OAPI_KEY_NUMPAD1);
			RESETKEY (kstate, OAPI_KEY_NUMPAD3);
			RESETKEY (kstate, OAPI_KEY_NUMPAD5);
		}

		//if ((auto_alt_status + auto_hvr_status) == 0)
		//{
		if (KEYDOWN (kstate, OAPI_KEY_ADD)) 
			{
			RESETKEY (kstate, OAPI_KEY_ADD);
			throttle_mode = 1;
			}

		if (KEYDOWN (kstate, OAPI_KEY_SUBTRACT)) 
			{
			RESETKEY (kstate, OAPI_KEY_SUBTRACT);
			throttle_mode = -1;
			}
		//}

		/////// pitch
		if (KEYDOWN (kstate, OAPI_KEY_NUMPAD2)) 
		{
		RESETKEY (kstate, OAPI_KEY_NUMPAD2);
		pitch_mode = 1;
		}

		if (KEYDOWN (kstate, OAPI_KEY_NUMPAD8)) 
		{
		RESETKEY (kstate, OAPI_KEY_NUMPAD8);
		pitch_mode = -1;
		}

		//// bank
		if (KEYDOWN (kstate, OAPI_KEY_NUMPAD4)) 
		{
		RESETKEY (kstate, OAPI_KEY_NUMPAD4);
		bank_mode = 1;
		}

		if (KEYDOWN (kstate, OAPI_KEY_NUMPAD6)) 
		{
		RESETKEY (kstate, OAPI_KEY_NUMPAD6);
		bank_mode = -1;
		}

		//// pitch trim
		/*if (KEYDOWN (kstate, OAPI_KEY_N)) 
		{
		RESETKEY (kstate, OAPI_KEY_N);
		rotor_trim = rotor_trim + step_time;
		if (rotor_trim > 1)rotor_trim = 1;
		}

		if (KEYDOWN (kstate, OAPI_KEY_M)) 
		{
		RESETKEY (kstate, OAPI_KEY_M);
		rotor_trim = rotor_trim - step_time;
		if (rotor_trim < 0)rotor_trim = 0;
		}*/

		/*if (KEYDOWN (kstate, OAPI_KEY_LEFT)) 
		{
		spot_az_status = -1;
		}

		if (KEYDOWN (kstate, OAPI_KEY_RIGHT)) 
		{
		spot_az_status = 1;
		}

		if (KEYDOWN (kstate, OAPI_KEY_UP)) 
		{
		spot_alt_status = -1;
		}

		if (KEYDOWN (kstate, OAPI_KEY_DOWN)) 
		{
		spot_alt_status = 1;
		}*/

	}

	return 0;
}


void titan_dragonfly::clbkPostCreation()
{
	if (eng_status == 2) 
	{
	SetMeshVisibilityMode(1,MESHVIS_NEVER); // 1
	SetMeshVisibilityMode(2,MESHVIS_ALWAYS); // 2
	}
	
	

	///////////// sound
	MyID=ConnectToOrbiterSoundDLL(GetHandle());

	ReplaceStockSound(MyID,"Sound\\_CustomVesselsSounds\\titan_dragonfly\\null.wav",REPLACE_RCS_THRUST_ATTACK);
	ReplaceStockSound(MyID,"Sound\\_CustomVesselsSounds\\titan_dragonfly\\null.wav",REPLACE_RCS_THRUST_SUSTAIN);
	ReplaceStockSound(MyID,"Sound\\_CustomVesselsSounds\\titan_dragonfly\\null.wav",REPLACE_MAIN_THRUST);
//	RequestLoadVesselWave(MyID,STRT,"Sound\\_CustomVesselsSounds\\titan_dragonfly\\demarragehelico.wav",BOTHVIEW_FADED_FAR);
	RequestLoadVesselWave(MyID,MOTR,"Sound\\_CustomVesselsSounds\\titan_dragonfly\\drone.wav",BOTHVIEW_FADED_FAR);
	SoundOptionOnOff(MyID,PLAYRADIOATC,FALSE);
	SoundOptionOnOff(MyID,PLAYCOUNTDOWNWHENTAKEOFF,FALSE);
	SoundOptionOnOff(MyID,PLAYLANDINGANDGROUNDSOUND,FALSE);
	SoundOptionOnOff(MyID,PLAYCABINAIRCONDITIONING,FALSE);
	SoundOptionOnOff(MyID,PLAYCABINRANDOMAMBIANCE,FALSE);
	SoundOptionOnOff(MyID,DISPLAYTIMER,FALSE);
	SoundOptionOnOff(MyID,PLAYDOCKLANDCLEARANCE,FALSE);
	SoundOptionOnOff(MyID,PLAYATTITUDETHRUST,FALSE);

	SetAttitudeMode (RCS_ROT);

}



void titan_dragonfly::clbkPreStep(double SimT, double SimDT, double MJD)
{
	step_time = SimDT;

	if(eng_status == 0)DoBatteryCharge();

	// get joystick input
	js_pu = GetManualControlLevel(THGROUP_ATT_PITCHUP,MANCTRL_ANYMODE,MANCTRL_JOYSTICK);
	js_pd = GetManualControlLevel(THGROUP_ATT_PITCHDOWN,MANCTRL_ANYMODE,MANCTRL_JOYSTICK);
	js_br = GetManualControlLevel(THGROUP_ATT_BANKRIGHT,MANCTRL_ANYMODE,MANCTRL_JOYSTICK);
	js_bl = GetManualControlLevel(THGROUP_ATT_BANKLEFT,MANCTRL_ANYMODE,MANCTRL_JOYSTICK);
	js_yl = GetManualControlLevel(THGROUP_ATT_YAWLEFT,MANCTRL_ANYMODE,MANCTRL_JOYSTICK);
	js_yr = GetManualControlLevel(THGROUP_ATT_YAWRIGHT,MANCTRL_ANYMODE,MANCTRL_JOYSTICK);
	//sprintf(oapiDebugString(), "js_pu %0.2f js_pd %0.2f js_br %0.2f js_bl %0.2f js_yl %0.2f js_yr %0.2f", js_pu, js_pd, js_br, js_bl, js_yl, js_yr);
	js_pitch_level = js_pu - js_pd;
	js_bank_level = js_bl - js_br;
	js_yaw_level = js_yr - js_yl;

	// set  thrust dir 
	SetMainTh(); 

	//////// do micro throttle
	if((auto_alt_status + auto_hvr_status) == 0)
	{
	if(throttle_mode != 0) 
		{
		dthrottle = (throttle_mode * step_time * 0.2);
		double throt_lvl = GetThrusterLevel(th_main[0]) + dthrottle;
		if(throt_lvl < 0)throt_lvl = 0;
		else if(throt_lvl > 1)throt_lvl = 1;
		SetThrusterLevel(th_main[0], throt_lvl);
		}
	else dthrottle = 0;
	}
	


	///// auto functions
	if(auto_alt_status == 1)DoHoldAlt();
	else if(auto_hvr_status == 1)DoHoldPos();

	///// reset for next frame
	pitch_mode = 0;
	bank_mode = 0;
	if (throttle_mode == 0)dthrottle = 0;
	throttle_mode = 0;

	//////// sound ////////////////////////////////////
	if (eng_status > 0)
	{
	double thr_lev = GetThrusterLevel(th_main[0]);
	int motr_level = (int) (thr_lev*100) + 150;
	if(oapiCameraInternal() == true) motr_level = motr_level - 100;
	int freq_lvl = (int) ((thr_lev * 4025) + 35050);
	PlayVesselWave(MyID, MOTR, LOOP, motr_level, freq_lvl);
	}
	else StopVesselWave(MyID, MOTR);


	///// anims ///////////////////////////
	double thmain_lvl = GetThrusterLevel(th_main[0]);
	if(thmain_lvl == 0)
	{
		if (eng_status == 2)
		{
		SetMeshVisibilityMode(1,MESHVIS_ALWAYS); // 1
		SetMeshVisibilityMode(2,MESHVIS_NEVER); // 2
		}
		eng_status = 0;
	}
	else if (thmain_lvl < 0.25)
	{
		if (eng_status == 2)
		{
		SetMeshVisibilityMode(1,MESHVIS_ALWAYS); // 1
		SetMeshVisibilityMode(2,MESHVIS_NEVER); // 2
		}
		eng_status = 1;
	}
	else 
	{
		if (eng_status < 2)
		{
		SetMeshVisibilityMode(1,MESHVIS_NEVER); // 1
		SetMeshVisibilityMode(2,MESHVIS_ALWAYS); // 2
		}
		eng_status = 2;
	}

	if (eng_status == 0)
	{
		SetAnimation(anim_rtr, rtr_proc);
		//SetThrusterGroupLevel(THGROUP_MAIN, 0);
	}
	else if (eng_status == 1)
	{
		double max_rps = 10;
		drot = thmain_lvl * 4;
		if(drot > 1)drot = 1;
		double abs_drot = drot * max_rps;
		
		if(abs_drot == max_rps)
		{
			eng_status = 2;
			SetMeshVisibilityMode(1,MESHVIS_NEVER); // 1
			SetMeshVisibilityMode(2,MESHVIS_ALWAYS); // 2
		}
		rtr_proc = rtr_proc + (abs_drot * step_time);
		int rtr_proc_int = (int)(rtr_proc);
		rtr_proc = rtr_proc - rtr_proc_int; // always between 0...1

		SetAnimation(anim_rtr, rtr_proc);

		//SetThrusterGroupLevel(THGROUP_MAIN, 0);
	}
	else if (eng_status == 2)
	{
		double max_rps = 6;
		double abs_drot = step_time * max_rps;
		
		blur_proc = blur_proc + abs_drot;
		int blur_proc_int = (int)(blur_proc);
		blur_proc = blur_proc - blur_proc_int; // always between 0...1

		SetAnimation(anim_blur, blur_proc);
	}
	//else if (eng_status == 3)
	//{
	//	double max_rps = 6;
	//	drot = drot - (step_time * 0.25);
	//	if(drot < 0)drot = 0;
	//	double abs_drot = drot * max_rps;
	//	
	//	if(abs_drot == 0)eng_status = 0;
	//	
	//	rtr_proc = rtr_proc + (abs_drot * step_time);
	//	int rtr_proc_int = (int)(rtr_proc);
	//	rtr_proc = rtr_proc - rtr_proc_int; // always between 0...1
	//	if(rtr_proc < 0)rtr_proc = rtr_proc + 1;// always between 0...1

	//	SetAnimation(anim_rtr, rtr_proc);
	//	//SetThrusterGroupLevel(THGROUP_MAIN, 0);
	//}

	if(autoland_status > 0)DoAutoLand();
	

	////////////////////// skis ///////////////


	if(ski_status == 2) ski_proc = 1;
	else if(ski_status == 1)
	{
		ski_proc = ski_proc - step_time;
		if(ski_proc < 0)
		{
		 ski_proc = 0;
		 ski_status = 0;
		}
	}
	else if(ski_status == 0)ski_proc = 0;

	SetAnimation(anim_ski, ski_proc);

	///////////////////////////////////// hga /////////////////////////

	if(ant_status > 0)DoHGA();

	///////////////land or boom ///////////////////

	if(GetAttachmentStatus(hAtt) == NULL)
	{
	if (boom_sw == 0 && GetAltitude(ALTMODE_GROUND) < 0.5)
		{
		VECTOR3 ves_ypos_hrz;
		HorizonRot(_V(0,1,0), ves_ypos_hrz);
		double vert_angle_cos = dotp(_V(0,1,0),ves_ypos_hrz);
		if (vert_angle_cos < 0.25) boom_sw = 1;

		VECTOR3 grnd_spd;
		GetGroundspeedVector(FRAME_HORIZON, grnd_spd);
		if (grnd_spd.y < -50)boom_sw = 1;

		}
	}
	if (boom_sw > 0 && boom_sw < 2) DoBoom();

	if (GetPropellantMass(hProp) == 0)boom_sw = 2;

	
}

void titan_dragonfly::clbkPostStep (double simt, double simdt, double mjd)
{
	if(GroundContact() == true && GetGroundspeed() > 0) 
	{
		VECTOR3 ves_ypos_hrz;
		HorizonRot(_V(0,1,0), ves_ypos_hrz);
		double vert_angle_cos = dotp(_V(0,1,0),ves_ypos_hrz);
		if (vert_angle_cos > 0.98 && GetGroundspeed() < 0.1)
		{
			if(GetThrusterGroupLevel(THGROUP_MAIN) == 0)
			{
			VESSELSTATUS2 vs;
			memset (&vs, 0, sizeof(vs));
			vs.version = 2;
			vs.rbody = GetGravityRef();
			vs.status = 1;
			OBJHANDLE hVessel = GetHandle();
			double heading;
			oapiGetHeading(hVessel,&heading);
			double lng, lat, rad;
			GetEquPos(lng, lat, rad);
			vs.surf_lat = lat;
			vs.surf_lng = lng;
			vs.surf_hdg = heading;
			vs.arot.x = 10;
			DefSetStateEx (&vs);
			}
		}
	}
}

void titan_dragonfly::DoBatteryCharge (void)
{
	double totfuel = (step_time/864000) + GetPropellantMass(hProp);
	if(totfuel > FUEL_MASS) SetPropellantMass(hProp, FUEL_MASS);
	else SetPropellantMass(hProp, totfuel);
}

void titan_dragonfly::DoHGA (void)
{
	double Earth_alt_angle = 0;
	double Earth_az_angle = 0;
	OBJHANDLE hEarth = oapiGetObjectByName("Earth");
	VECTOR3 Earth_pos = _V(0,0,0);
	oapiGetGlobalPos(hEarth, &Earth_pos);
	VECTOR3 Earth_pos_loc  = _V(0,0,0);
	Global2Local(Earth_pos, Earth_pos_loc);
	normalise(Earth_pos_loc);

	double drotmax = step_time*0.1;

	if(ant_status == 1)
	{
		ant_alt_proc = ant_alt_proc + (step_time * 0.1);
		if(ant_alt_proc > 0.5)
		{
			ant_alt_proc = 0.5;
			ant_status = 2;
		}
	}
	else if(ant_status == 2)
	{
		Earth_alt_angle = PI05 - acos(dotp(Earth_pos_loc, _V(0,1,0)));
		double tgt_alt_proc = 0.5 + (Earth_alt_angle * (1/PI));
		if(tgt_alt_proc < 0.5) tgt_alt_proc = 0.5;
		if(tgt_alt_proc > 1)tgt_alt_proc = 1;

		
		if(abs(tgt_alt_proc - ant_alt_proc) < drotmax)ant_alt_proc = tgt_alt_proc;
		else if(tgt_alt_proc > ant_alt_proc)
		{
			ant_alt_proc = ant_alt_proc + drotmax;
			if(ant_alt_proc > 1)ant_alt_proc = 1;
		}
		else
		{
			ant_alt_proc = ant_alt_proc - drotmax;
			if(ant_alt_proc < 0.5)ant_alt_proc = 0.5;
		}

	//////////////////// az
	Earth_az_angle = atan2(-Earth_pos_loc.x,-Earth_pos_loc.z);
	if(Earth_az_angle < 0)Earth_az_angle = Earth_az_angle + PI2;
	double tgt_az_proc = Earth_az_angle * (1/PI2);
	if(abs(ant_az_proc - tgt_az_proc) < drotmax)ant_az_proc = tgt_az_proc;
	else
	{
	double drot = tgt_az_proc - ant_az_proc;
	int dir = 1;
	if(drot > 0.5) dir = -1;
	else if(drot < -0.5) dir = 1;
	else if(drot < 0) dir = -1;

	ant_az_proc = ant_az_proc + (drotmax * dir);
	}
	if(ant_az_proc > 1) ant_az_proc = ant_az_proc - (int)(ant_az_proc);
	else if(ant_az_proc < 0) ant_az_proc = 1 + ant_az_proc - (int)(ant_az_proc);
	}
	else if(ant_status == 3)
	{
	ant_alt_proc = ant_alt_proc - drotmax;
	if(ant_alt_proc < 0)ant_alt_proc = 0;
	//////////
	if(ant_az_proc < drotmax)ant_az_proc = 0;
	else
		{
		int dir = 1;
		if(ant_az_proc < 0.5) dir = -1;

		ant_az_proc = ant_az_proc + (drotmax * dir);
		if(ant_az_proc > 1) ant_az_proc = ant_az_proc - (int)(ant_az_proc);
		else if(ant_az_proc < 0) ant_az_proc = 1 + ant_az_proc - (int)(ant_az_proc);
		}

	if((ant_az_proc + ant_alt_proc) == 0)ant_status = 0; 
	}

	SetAnimation(anim_ant_az, ant_az_proc);
	SetAnimation(anim_ant_alt, ant_alt_proc);

}

void titan_dragonfly::SetMainTh (void)
{
	double stp = GetAtmPressure ();
	if(stp < 1)stp = 1;
	double alt_factor = pow((stp / 148900),1.25);
	double new_max_th = MAIN_TH*alt_factor;
	double new_isp = MAIN_ISP*alt_factor;

	SetThrusterMax0(th_main[0], new_max_th);
	SetThrusterIsp(th_main[0], new_isp);

	double max_pitch_angle = (30*RAD);
	double max_bank_angle = (30*RAD);
	VECTOR3 th_axis = _V(0,1,0);
	VECTOR3 pitch_axis = _V(-1,0,0);
	VECTOR3 bank_axis = _V(0,0,1);

//	double rotor_trim_angle = -rotor_trim * (10*RAD);
//	MATRIX3 rtnmtx = rotm(pitch_axis, rotor_trim_angle);
//	th_axis = mul(rtnmtx, th_axis);

	if (pitch_mode == 0) // return to centre
	{
		if (pitch_level < 0) 
			{
			pitch_level = pitch_level + step_time;
			if (pitch_level > 0)pitch_level = 0;
			}
		else if (pitch_level > 0) 
			{
			pitch_level = pitch_level - step_time;
			if (pitch_level < 0)pitch_level = 0;
			}
	}
	else // pitch control active
	{
		pitch_level = pitch_level + (pitch_mode * step_time);
		if (pitch_level > 1)pitch_level = 1;
		else if (pitch_level < -1)pitch_level = -1;
	}

	double pitch_angle = max_pitch_angle * (pitch_level + js_pitch_level);
	MATRIX3 rtnmtx = rotm(pitch_axis, pitch_angle);
	th_axis = mul(rtnmtx, th_axis);

	/////////////////////////

	if (bank_mode == 0)
	{
		if (bank_level < 0) 
			{
			bank_level = bank_level + step_time;
			if (bank_level > 0)bank_level = 0;
			}
		else if (bank_level > 0) 
			{
			bank_level = bank_level - step_time;
			if (bank_level < 0)bank_level = 0;
			}
	}
	else 
	{
		bank_level = bank_level + (bank_mode * step_time);
		if (bank_level > 1)bank_level = 1;
		else if (bank_level < -1)bank_level = -1;
	}
	bank_axis = _V(0, -th_axis.z, th_axis.y);

	double bank_angle = max_bank_angle * (bank_level + js_bank_level);
	rtnmtx = rotm(bank_axis, bank_angle);
	th_axis = mul(rtnmtx, th_axis);

	SetThrusterDir(th_main[0], th_axis);

	//sprintf(oapiDebugString(), "th_axis %0.3f %0.3f %0.3f rotor_trim_angle %0.3f", th_axis.x,th_axis.y,th_axis.z,rotor_trim_angle*DEG);
}



bool titan_dragonfly::clbkDrawHUD (int mode, const HUDPAINTSPEC *hps, oapi::Sketchpad *skp)
{
	// draw the default HUD
	VESSEL3::clbkDrawHUD (mode, hps, skp);

	int s = hps->H;
	double d = (s*0.00130208);
	int sw = ((hps->W));
	int lw =  (int)(16*sw/1024);
	int lwoffset = sw - (18*lw);
	int hlw = (int) (lw/2);


	int roxl = 0;
	int royl = 0;

	double ds = s;
	double dsw = sw;
	double sc_ratio = ds/dsw;


	if (sc_ratio < 0.7284)
	{
	roxl = (lw*10);
	royl = (int) (-88*d);
	}



//	int wd = (int)(152*d);
//	int wc = (int)(168*d);
	int w0 = (int)(184*d);
	int w1 = (int) (200*d);
	int w2 = (int)(216*d);
	int w3 = (int)(232*d);
	int w4 = (int)(248*d);
//	int w5 = (int)(264*d);
//	int w6 = (int)(280*d);
//	int w7 = (int)(296*d);
//	int w8 = (int)(312*d);
//	int w9 = (int)(328*d);
//	int w10 = (int)(344*d);
//	int w11 = (int)(360*d);

	/*{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];

	if(hook_status == 0)sprintf(abuf,"G = Deploy Grapple");
	else if(hook_status == 1)sprintf(abuf,"G = Stow Grapple");
	else if(hook_status == 2)
		{
		if(grapple_status == 3)sprintf(abuf,"G = Reset Grapple");
		else sprintf(abuf,"G = Stow Grapple");
		}
	else if(hook_status == 3)
		{
		if(grapple_status == 3)sprintf(abuf,"G = Cancel Reset Grapple");
		else sprintf(abuf,"G = Stow Grapple");
		}
	skp->Text ((10+roxl), (wc+royl), abuf,strlen(abuf));
	}*/

	{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];
	sprintf(abuf,"G = Deploy Skids");
	if(ski_status == 2)skp->Text ((10+roxl), (w0+royl), abuf,strlen(abuf));
	}

	{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];
	sprintf(abuf,"M = Deploy/Stow HGA");
	skp->Text ((10+roxl), (w1+royl), abuf,strlen(abuf));
	}

	{
	if(auto_hvr_status == 0) skp->SetTextColor(0x0066FF66);
	else skp->SetTextColor(0x0066FFFF);
	char abuf[256];
	sprintf(abuf,"V = AutoHover On/Off");
	skp->Text ((10+roxl), (w2+royl), abuf,strlen(abuf));
	}

	{
	if(auto_alt_status == 0) skp->SetTextColor(0x0066FF66);
	else skp->SetTextColor(0x0066FFFF);
	char abuf[256];
	sprintf(abuf,"B = Hold Alt. On/Off");
	skp->Text ((10+roxl), (w3+royl), abuf,strlen(abuf));
	}

	{
	if(autoland_status == 0) skp->SetTextColor(0x0066FF66);
	else skp->SetTextColor(0x0066FFFF);
	char abuf[256];
	sprintf(abuf,"N = AutoLand On/Off");
	skp->Text ((10+roxl), (w4+royl), abuf,strlen(abuf));
	}



	return true;
}

void titan_dragonfly::DoAutoLand (void)
{
	double altg = GetAltitude(ALTMODE_GROUND);

	if(altg < 1.0)
	{
		double th1_level = GetThrusterLevel(th_main[0]) - (step_time*0.25);
		SetThrusterLevel(th_main[0], th1_level);
		
		if(th1_level < 0)
		{
		SetThrusterLevel(th_main[0], 0);
		SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT,0);
		SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT,0);
		autoland_status = 0;
		return;
		}
		return;
	}

	VECTOR3 grd_spd_vec_hrz;
	GetGroundspeedVector(FRAME_HORIZON, grd_spd_vec_hrz);

	VECTOR3 weight_vec;
	GetWeightVector(weight_vec);
	double weight = length(weight_vec);
	weight_vec = _V(0,-weight,0);

	VECTOR3 drag_vector;
	GetDragVector(drag_vector);

	double ground_spd = length(_V(grd_spd_vec_hrz.x,0,grd_spd_vec_hrz.z));
	double stop_timexz = ground_spd*2;
	if(stop_timexz < 2)stop_timexz = 2;
	double stop_timey = abs(grd_spd_vec_hrz.y)*2;
	if(stop_timey < 1)stop_timey = 2;
	double mass = GetMass();
	double angle_limit = 10*RAD;
	double y_vel_off = -20;

	double hdg;
	oapiGetHeading (GetHandle(), &hdg);
	double x_vel_off = grd_spd_vec_hrz.x;
	double z_vel_off = grd_spd_vec_hrz.z;

	if(altg < 1000)
	{
		y_vel_off = -10;
		x_vel_off = 5*(sin(hdg));
		z_vel_off = 5*(cos(hdg));
	}
	if(altg < 500)
	{
		y_vel_off = -5;
		x_vel_off = 5*(sin(hdg));
		z_vel_off = 5*(cos(hdg));
	}
	if(altg < 100)
	{
		y_vel_off = -2;
		x_vel_off = 2*(sin(hdg));
		z_vel_off = 2*(cos(hdg));
	}
	if(altg < 50)
	{
		y_vel_off = -1;
		x_vel_off = 1*(sin(hdg));
		z_vel_off = 1*(cos(hdg));
	}
	if(altg < 20)
	{
		y_vel_off = -0.5;
		x_vel_off = 0;
		z_vel_off = 0;
	}
	if(altg < 2)
	{
		y_vel_off = -0.25;
		x_vel_off = 0;
		z_vel_off = 0;
	}

	double req_acc_x = -(grd_spd_vec_hrz.x-x_vel_off)*(1/stop_timexz);
	double req_acc_y = -(grd_spd_vec_hrz.y-y_vel_off)*(1/stop_timey);
	//if(req_acc_y < 0)req_acc_y = 0;
	double req_acc_z = -(grd_spd_vec_hrz.z-z_vel_off)*(1/stop_timexz);

	VECTOR3 req_force_x = _V(req_acc_x*mass,0,0);
	VECTOR3 req_force_y = _V(0,req_acc_y*mass,0);
	VECTOR3 req_force_z = _V(0,0,req_acc_z*mass);

	VECTOR3 req_thrust_hrz = req_force_x + req_force_y + req_force_z - weight_vec - drag_vector;
	if(req_thrust_hrz.y < 0)req_thrust_hrz.y = 0;

	double max_th = GetThrusterMax(th_main[0]);

	double th_level = length(req_thrust_hrz)/max_th;

	

	VECTOR3 req_thrust_loc;
	HorizonInvRot(req_thrust_hrz, req_thrust_loc);
	normalise(req_thrust_loc);

	double req_angle = acos(dotp(_V(0,1,0), req_thrust_loc));
	if (req_angle > angle_limit)
	{
		VECTOR3 th_rot_axis = crossp(_V(0,1,0),req_thrust_loc);
		normalise(th_rot_axis);
		MATRIX3 rtnmtx = rotm(th_rot_axis, angle_limit);//???????????????
		req_thrust_loc = mul(rtnmtx, _V(0,1,0));
		normalise(req_thrust_loc);
		//sprintf(oapiDebugString(), "rotaxis %0.3f %0.3f %0.3f", th_rot_axis.x, th_rot_axis.y, th_rot_axis.z);
	}

	////////////////////// user input //////////////

	//////////// pitch n bank
	VECTOR3 pitch_axis = _V(-1,0,0);
	VECTOR3 bank_axis = _V(0,0,1);

	///// pitch
	if (pitch_mode == 0) // return to centre
	{
		if (pitch_level < 0) 
			{
			pitch_level = pitch_level + step_time;
			if (pitch_level > 0)pitch_level = 0;
			}
		else if (pitch_level > 0) 
			{
			pitch_level = pitch_level - step_time;
			if (pitch_level < 0)pitch_level = 0;
			}
	}
	else // pitch control active
	{
		pitch_level = pitch_level + (pitch_mode * step_time);
		if (pitch_level > 1)pitch_level = 1;
		else if (pitch_level < -1)pitch_level = -1;
	}

	double user_pitch_angle = (js_pitch_level + pitch_level)*(10*RAD);
	MATRIX3 rtnmtx2 = rotm(pitch_axis, user_pitch_angle);
	req_thrust_loc = (rtnmtx2, req_thrust_loc);

	//// bank

	if (bank_mode == 0)
	{
		if (bank_level < 0) 
			{
			bank_level = bank_level + step_time;
			if (bank_level > 0)bank_level = 0;
			}
		else if (bank_level > 0) 
			{
			bank_level = bank_level - step_time;
			if (bank_level < 0)bank_level = 0;
			}
	}
	else 
	{
		bank_level = bank_level + (bank_mode * step_time);
		if (bank_level > 1)bank_level = 1;
		else if (bank_level < -1)bank_level = -1;
	}

	double user_bank_angle = (js_bank_level + bank_level)*(10*RAD);
	rtnmtx2 = rotm(bank_axis, user_bank_angle);
	req_thrust_loc = mul(rtnmtx2, req_thrust_loc);

	///////////// throttle


	if(throttle_mode != 0) 
	{
		dthrottle = (throttle_mode * step_time * 0.1);
		double throt_lev = GetThrusterLevel(th_main[0]);
		throt_lev = throt_lev + dthrottle;
		if (throt_lev > 1)throt_lev = 1;
		else if(throt_lev < 0.15)throt_lev = 0.15;
		SetThrusterLevel(th_main[0],throt_lev);
	}
	else 
	{
		if (th_level > 1)th_level = 1;
		else if(th_level < 0.15)th_level = 0.15;
		SetThrusterLevel(th_main[0], th_level);
	}

	SetThrusterDir(th_main[0], req_thrust_loc);

	//// y axis //////////////////////
	VECTOR3 ang_vel;
	GetAngularVel(ang_vel);

	double drot_y = step_time*ang_vel.y;
	double max_y_torque = 1*YAW_TH;
	double max_y_acc = max_y_torque/(mass*1);
	double max_y_dvrot = max_y_acc*step_time;
	double req_level = ang_vel.y/max_y_dvrot;
	if(req_level > 0.5)req_level = 0.5;
	else if(req_level < -0.5)req_level = -0.5;
	if(req_level > 0)
	{
	SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT,0);
	SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT,req_level);
	}
	else if(req_level < 0)
	{
	SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT,-req_level);
	SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT,0);
	}
	else
	{
	SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT,0);
	SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT,0);
	}



}

void titan_dragonfly::DoHoldPos(void)
{

	VECTOR3 grd_spd_vec_hrz;
	GetGroundspeedVector(FRAME_HORIZON, grd_spd_vec_hrz);

	VECTOR3 weight_vec;
	GetWeightVector(weight_vec);
	double weight = length(weight_vec);
	weight_vec = _V(0,-weight,0);

	VECTOR3 drag_vector;
	GetDragVector(drag_vector);

	double ground_spd = GetGroundspeed();
	double stop_time = ground_spd/1;
	if(stop_time < 1)stop_time = 1;
	double mass = GetMass();
	double angle_limit = 30*RAD;

	double req_acc_x = -grd_spd_vec_hrz.x/stop_time;
	double req_acc_y = -grd_spd_vec_hrz.y/(stop_time*0.2);
	double req_acc_z = -grd_spd_vec_hrz.z/stop_time;

	VECTOR3 req_force_x = _V(req_acc_x*mass,0,0);
	VECTOR3 req_force_y = _V(0,req_acc_y*mass,0);
	VECTOR3 req_force_z = _V(0,0,req_acc_z*mass);

	VECTOR3 req_thrust_hrz = req_force_x + req_force_y + req_force_z - weight_vec;// - drag_vector; 

	double max_th = GetThrusterMax(th_main[0]);

	double th_level = length(req_thrust_hrz)/max_th;
	

	VECTOR3 req_thrust_loc;
	HorizonInvRot(req_thrust_hrz, req_thrust_loc);
	normalise(req_thrust_loc);

	double req_angle = acos(dotp(_V(0,1,0), req_thrust_loc));
	if (req_angle > angle_limit)
	{
		VECTOR3 th_rot_axis = crossp(_V(0,1,0),req_thrust_loc);
		normalise(th_rot_axis);
		MATRIX3 rtnmtx = rotm(th_rot_axis, angle_limit);//???????????????
		req_thrust_loc = mul(rtnmtx, _V(0,1,0));
		normalise(req_thrust_loc);
		//sprintf(oapiDebugString(), "rotaxis %0.3f %0.3f %0.3f", th_rot_axis.x, th_rot_axis.y, th_rot_axis.z);
	}

	////////////////////// user input //////////////

	//////////// pitch n bank
	VECTOR3 pitch_axis = _V(-1,0,0);
	VECTOR3 bank_axis = _V(0,0,1);

	///// pitch
	if (pitch_mode == 0) // return to centre
	{
		if (pitch_level < 0) 
			{
			pitch_level = pitch_level + step_time;
			if (pitch_level > 0)pitch_level = 0;
			}
		else if (pitch_level > 0) 
			{
			pitch_level = pitch_level - step_time;
			if (pitch_level < 0)pitch_level = 0;
			}
	}
	else // pitch control active
	{
		pitch_level = pitch_level + (pitch_mode * step_time);
		if (pitch_level > 1)pitch_level = 1;
		else if (pitch_level < -1)pitch_level = -1;
	}

	double user_pitch_angle = (js_pitch_level + pitch_level)*(10*RAD);
	MATRIX3 rtnmtx2 = rotm(pitch_axis, user_pitch_angle);
	req_thrust_loc = (rtnmtx2, req_thrust_loc);

	//// bank

	if (bank_mode == 0)
	{
		if (bank_level < 0) 
			{
			bank_level = bank_level + step_time;
			if (bank_level > 0)bank_level = 0;
			}
		else if (bank_level > 0) 
			{
			bank_level = bank_level - step_time;
			if (bank_level < 0)bank_level = 0;
			}
	}
	else 
	{
		bank_level = bank_level + (bank_mode * step_time);
		if (bank_level > 1)bank_level = 1;
		else if (bank_level < -1)bank_level = -1;
	}

	double user_bank_angle = (js_bank_level + bank_level)*(10*RAD);
	rtnmtx2 = rotm(bank_axis, user_bank_angle);
	req_thrust_loc = mul(rtnmtx2, req_thrust_loc);

	///////////// throttle


	if(throttle_mode != 0) 
	{
		dthrottle = (throttle_mode * step_time * 0.1);
		double throt_lev = GetThrusterLevel(th_main[0]);
		throt_lev = throt_lev + dthrottle;
		if (throt_lev > 1)throt_lev = 1;
		else if(throt_lev < 0.15)throt_lev = 0.15;
		SetThrusterLevel(th_main[0],throt_lev);
	}
	else 
	{
		if (th_level > 1)th_level = 1;
		else if(th_level < 0.15)th_level = 0.15;
		SetThrusterLevel(th_main[0], th_level);
	}

	SetThrusterDir(th_main[0], req_thrust_loc);

	//// y axis //////////////////////
	VECTOR3 ang_vel;
	GetAngularVel(ang_vel);

	double drot_y = step_time*ang_vel.y;
	double max_y_torque = 1*YAW_TH;
	double max_y_acc = max_y_torque/(mass*1);
	double max_y_dvrot = max_y_acc*step_time;
	double req_level = ang_vel.y/max_y_dvrot;
	if(req_level > 0.5)req_level = 0.5;
	else if(req_level < -0.5)req_level = -0.5;
	if(req_level > 0)
	{
	SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT,0);
	SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT,req_level);
	}
	else if(req_level < 0)
	{
	SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT,-req_level);
	SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT,0);
	}
	else
	{
	SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT,0);
	SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT,0);
	}

}

void titan_dragonfly::DoHoldAlt(void)
{
	VECTOR3 th_axis_loc;
	GetThrusterDir(th_main[0],th_axis_loc);
	VECTOR3 th_axis_hrz;
	HorizonRot(th_axis_loc, th_axis_hrz);
	normalise(th_axis_hrz);

	VECTOR3 grd_spd_hrz;
	GetGroundspeedVector(FRAME_HORIZON, grd_spd_hrz);
	double stop_time = abs(grd_spd_hrz.y)/5;
	if(stop_time < 0.1)stop_time = 0.1;
	double req_acc = -grd_spd_hrz.y/stop_time;
	double mass = GetMass();
	double vvel_force = mass*req_acc;

	VECTOR3 weight_vect;
	GetWeightVector(weight_vect);
	double weight = length(weight_vect);
	double req_vert_force = weight + vvel_force;

	double req_total_force = req_vert_force/th_axis_hrz.y;

	double th_lev = req_total_force/GetThrusterMax(th_main[0]);

	if(throttle_mode != 0) 
	{
		dthrottle = (throttle_mode * step_time * 0.1);
		double throt_lev = GetThrusterLevel(th_main[0]);
		throt_lev = throt_lev + dthrottle;
		if (throt_lev > 1)throt_lev = 1;
		else if(throt_lev < 0.15)throt_lev = 0.15;
		if (th_axis_hrz.y > 0) SetThrusterLevel(th_main[0], throt_lev);
	}
	else 
	{
		if (th_lev > 1)th_lev = 1;
		else if(th_lev < 0.15)th_lev = 0.15;
		if (th_axis_hrz.y > 0) SetThrusterLevel(th_main[0], th_lev);
	}


}

void titan_dragonfly::DoBoom(void)
{
//if(boom_sw == 1)
//{
VESSELSTATUS2 vs;
	memset (&vs, 0, sizeof(vs));
	vs.version = 2;
	GetStatusEx(&vs);
	OBJHANDLE hVessel = GetHandle();
	double longitude, latitude, radius;
	double heading;
	oapiGetEquPos(hVessel,&longitude,&latitude,&radius);
	oapiGetHeading(hVessel,&heading);
	vs.vrot.x = 1.43;
	vs.surf_lat = latitude;
	vs.surf_lng = longitude;
	vs.surf_hdg = heading;
	vs.status = 1;
	DefSetStateEx (&vs);

	eng_status = 0;
	drot = 0;
//}
	//SetMeshVisibilityMode(0,MESHVIS_NEVER);
	//SetMeshVisibilityMode(1,MESHVIS_NEVER);

	boom_sw = boom_sw + step_time;
	if (boom_sw < 2)
	{
		SetThrusterLevel(th_boom[0],1);
		SetThrusterLevel(th_boom[1],1);
	}
	else
	{
		boom_sw = 2;
		SetThrusterLevel(th_boom[0],0);
		SetThrusterLevel(th_boom[1],0);
		SetPropellantMass(hProp, 0);
	}
}