// ==============================================================
//                 ORBITER MODULE: titan_dragonfly_backshell
//                  
//
// titan_dragonfly_backshell.cpp
// Control module for titan_dragonfly_backshell vessel class
//
// 
// ==============================================================

#define STRICT
#define ORBITER_MODULE

#include "orbitersdk.h"
#include "OrbiterSoundSDK40.h"
//#include <cstdio>
//#include <cmath>
//#include <string>
#include <stdio.h>
#include <math.h>

// ==============================================================
// Some vessel parameters
// ==============================================================


const double CHUTE_OPERATING_SPEED = 1;
const double COLLAPSE_SPEED = 0.5;
const double BKSHL_MASS = 210;
const double HTSHLD_MASS = 220;
const double CRUISE_MASS = 320;
const double TD_MASS = 450;
const double FUEL_MASS = 400;
const double RCS_ISP = 3000;
const double RCS_TH = 20;
const double drg_size = 2;
const double drg_ref = 3;
const double par_size = 15;
const double par_ref = 3;
const double cap_size = 4;
const double cap_ref = 0.1;
const VECTOR3 CRUISE_OFF = _V(0,0,-1.2);

#define BOOM	1

double stiffness = 3e4;// 6e4
double damping = 1e4; // 5e4

static TOUCHDOWNVTX tdvtx_geardown[7] = {
	{_V( 0,2, 0.1),     stiffness, damping, 0.9, 0.2},
	{_V(-1,-1.732, 0.1), stiffness, damping, 0.9, 0.2},
	{_V( 1,-1.732, 0.1), stiffness, damping, 0.9, 0.2},
	{_V( 0,-2, 0.1),     stiffness, damping, 0.9},
	{_V(-1, 1.732, 0.1), stiffness, damping, 0.9},
	{_V( 1, 1.732, 0.1), stiffness, damping, 0.9},
	{_V(0,0, -2),        stiffness, damping, 0.9}
};


// Calculate lift coefficient [Cl] as a function of aoa (angle of attack) over -Pi ... Pi
// Implemented here as a piecewise linear function
void VLiftCoeff (VESSEL *v, double aoa, double M, double Re, void *context, double *cl, double *cm, double *cd)
{
	int i;
	const int nabsc = 9;
	static const double AOA[nabsc] = {-180*RAD,-60*RAD,-30*RAD, -2*RAD, 10*RAD,15*RAD,50*RAD,90*RAD,180*RAD};
	static const double CL[nabsc]  = {0,0,0,0,0,0,0,0,0};
	static const double CM[nabsc]  = {0,0,0,0,0,0,0,0,0};
	for (i = 0; i < nabsc-1 && AOA[i+1] < aoa; i++);
	double f = (aoa-AOA[i]) / (AOA[i+1]-AOA[i]);
	*cl = CL[i] + (CL[i+1]-CL[i]) * f;  // aoa-dependent lift coefficient
	*cm = 0.0;
	*cd = 0.95 + oapiGetInducedDrag (*cl, 1.5, 0.7) + oapiGetWaveDrag (M, 0.75, 1.0, 1.1, 0.04);
}

void HLiftCoeff (VESSEL *v, double beta, double M, double Re, void *context, double *cl, double *cm, double *cd)
{
	int i;
	const int nabsc = 8;
	static const double BETA[nabsc] = {-180*RAD,-135*RAD,-90*RAD,-45*RAD,45*RAD,90*RAD,135*RAD,180*RAD};
	static const double CL[nabsc]   = {0,0,0,0,0,0,0,0};
	for (i = 0; i < nabsc-1 && BETA[i+1] < beta; i++);
	*cl = CL[i] + (CL[i+1]-CL[i]) * (beta-BETA[i]) / (BETA[i+1]-BETA[i]);
	*cm = 0.0;
	*cd = 0.95 + oapiGetInducedDrag (*cl, 1.5, 0.7) + oapiGetWaveDrag (M, 0.75, 1.0, 1.1, 0.04);
}


class titan_dragonfly_backshell: public VESSEL3 {
public:
	
	titan_dragonfly_backshell (OBJHANDLE hVessel, int flightmodel);
	~titan_dragonfly_backshell ();

	void clbkSetClassCaps (FILEHANDLE cfg);
	void DefineAnimations (void);
	void clbkPreStep(double SimT, double SimDT, double MJD);
	void clbkPostStep (double simt, double simdt, double mjd);
	void clbkPostCreation();
	void clbkSaveState (FILEHANDLE scn);
	void clbkLoadStateEx (FILEHANDLE scn, void *status);
	int clbkConsumeBufferedKey(DWORD key, bool down, char *kstate);
	bool clbkDrawHUD (int mode, const HUDPAINTSPEC *hps, oapi::Sketchpad *skp);
	int MyID;
	

private:
	void JettisonCruise(void);
	void JettisonHeatshield(void);
	void JettisonTD(void);
	void CalcMass(void);
	void DoBoom(void);
	void GoToEDLAttitude(void);
	void DeleteRCS(void);
	void ApplySkin(void);
	void KillRCS(void);
	void SetAttachPoint(void);

	UINT anim_chute, anim_col, anim_drogue, anim_cable;
	double chute_proc, drogue_proc, col_proc, boom_proc, pstream_lvl, cable_proc;
	int chute_status, drogue_status, col_status, cruise_status, htshld_status, td_status, boom_status;
	int descent_status, lid_status, skin_status, attitude_status, cable_status;
	AIRFOILHANDLE hHA, hVA;
	ATTACHMENTHANDLE hTD_att, hFH_att;
	PSTREAM_HANDLE pHndl;
	THRUSTER_HANDLE th_rcs[8], th_group[2];
	PROPELLANT_HANDLE hProp;
	SURFHANDLE bs1, bs2, bs3, bs4;
	MESHHANDLE hMHeatshield, hMbackshell;
	double step_time;
	int landed_sw;
	//int report_sw;
};

titan_dragonfly_backshell::titan_dragonfly_backshell (OBJHANDLE hVessel, int flightmodel)
: VESSEL3 (hVessel, flightmodel)
{
}

titan_dragonfly_backshell::~titan_dragonfly_backshell ()
{
oapiDestroySurface(bs2);
oapiDestroySurface(bs4);
}

// Define animation sequences for moving parts
// --------------------------------------------------------------
void titan_dragonfly_backshell::DefineAnimations (void)
{
	////////// parachute

	anim_chute = CreateAnimation (0);
	// 1. chute reefed
	static UINT HiGainAnt1Grp[1] = {0};
	static MGROUP_SCALE HiGainAnt1 (3, HiGainAnt1Grp, 1, _V(0,0,-1.16),_V(0.01,0.01,0.005));
	AddAnimationComponent (anim_chute, 0.8, 1, &HiGainAnt1);


	
	// 2. unreef
	static UINT HiGainAnt2Grp[1] = {0};
	static MGROUP_SCALE HiGainAnt2 (3, HiGainAnt2Grp, 1, _V(0,0,-1.16),_V(0.5,0.5,1));
	AddAnimationComponent (anim_chute, 0, 0.5, &HiGainAnt2);

	// 3. collapse
	anim_col = CreateAnimation (0);

	static UINT HiGainAnt3Grp[1] = {0};
	static MGROUP_ROTATE HiGainAnt3 (3, HiGainAnt3Grp, 1, _V(0,0,-1.16),_V(1,0,0), 80*RAD);
	AddAnimationComponent (anim_col, 0, 0.8, &HiGainAnt3);

	static UINT HiGainAnt4Grp[1] = {0};
	static MGROUP_SCALE HiGainAnt4 (3, HiGainAnt4Grp, 1, _V(0,0,-0.1),_V(1,1,0.01));
	AddAnimationComponent (anim_col, 0.8, 1, &HiGainAnt4);

	/////////// drogue /////////////////

	anim_drogue = CreateAnimation (0);
	// 1. drogue reefed
	static UINT Drg1Grp[1] = {0};
	static MGROUP_SCALE Drg1 (4, Drg1Grp, 1, _V(0,0,-1.16),_V(0.01,0.01,0.005));
	AddAnimationComponent (anim_drogue, 0.8, 1, &Drg1);


	
	// 2. unreef
	static UINT Drg2Grp[1] = {0};
	static MGROUP_SCALE Drg2 (4, Drg2Grp, 1, _V(0,0,-1.16),_V(0.5,0.5,1));
	AddAnimationComponent (anim_drogue, 0, 0.5, &Drg2);

	////////// cables ////////////////////////////////////

	anim_cable = CreateAnimation (0);

	static UINT Cable1Grp[1] = {7};
	static MGROUP_SCALE Cable1 (0, Cable1Grp, 1, _V(0,0,-0.389),_V(1,1,10.204));
	AddAnimationComponent (anim_cable, 0, 1, &Cable1);

}



// ==============================================================
// titan_dragonfly_backshell class interface
// ==============================================================

// ==============================================================
// Overloaded callback functions
// ==============================================================

// --------------------------------------------------------------
// Set the capabilities of the vessel class
// --------------------------------------------------------------
void titan_dragonfly_backshell::clbkSetClassCaps (FILEHANDLE cfg)
{
	chute_proc = 1;
	chute_status = 2;
	drogue_proc = 1;
	drogue_status = 3;
	cruise_status = 0;
	col_proc = 0;
	col_status = 0;
	htshld_status = 0;
	td_status = 0;
	boom_proc = 0;
	pstream_lvl = 0;
	boom_status = 0;
	descent_status = 0;
	lid_status = 0;
	skin_status = 0;
	attitude_status = 0;
	step_time = 0;
	cable_proc = 0;
	cable_status = 0;
	landed_sw = 0;
	pHndl = NULL;
	//report_sw = 0;

	DefineAnimations ();
	// physical specs
	SetCW (4, 4, 2,2);
	SetRotDrag (_V(2,2,0.1));
	SetPMI (_V(1.4,1.4,1.5));
	SetCrossSections (_V(4,4,10));
	SetSize (2);
	CalcMass();
	SetCameraOffset (_V(0,0,0));
	SetTouchdownPoints ( tdvtx_geardown, 7);

	hMbackshell = oapiLoadMeshGlobal("titan_dragonfly/backshell");
	hMHeatshield = oapiLoadMeshGlobal("titan_dragonfly/heatshield");

	SetMeshVisibilityMode (AddMesh (hMbackshell), MESHVIS_EXTERNAL); //0
	SetMeshVisibilityMode (AddMesh (hMHeatshield), MESHVIS_EXTERNAL); //1
	SetMeshVisibilityMode (AddMesh ("titan_dragonfly/lid"), MESHVIS_EXTERNAL); //2
	SetMeshVisibilityMode (AddMesh ("titan_dragonfly/parachute"), MESHVIS_NEVER); //3
	SetMeshVisibilityMode (AddMesh ("titan_dragonfly/drogue"), MESHVIS_NEVER); //4
	SetMeshVisibilityMode (AddMesh ("titan_dragonfly/cruise", &CRUISE_OFF), MESHVIS_EXTERNAL); //5
	
	bs2 = oapiLoadTexture ("titan_dragonfly/bkshl_outer_burnt.dds", true);
	bs4 = oapiLoadTexture ("titan_dragonfly/hs_outer_burnt.dds", true);

	//////////////////////////////////////////

	hProp = CreatePropellantResource (FUEL_MASS, FUEL_MASS);

	
	// ***************** thruster definitions *******************

	th_rcs[0] = CreateThruster (_V( 1.913, 0, -0.61), _V( 0, 0,-1), RCS_TH, hProp, RCS_ISP);
	th_rcs[1] = CreateThruster (_V( 1.913, 0, -0.86), _V( 0, 0, 1), RCS_TH, hProp, RCS_ISP);
	th_rcs[2] = CreateThruster (_V( 1.913,-0.13, -0.73), _V( 0, 1, 0), RCS_TH, hProp, RCS_ISP);
	th_rcs[3] = CreateThruster (_V( 1.913, 0.13, -0.73), _V( 0,-1, 0), RCS_TH, hProp, RCS_ISP);

	th_rcs[4] = CreateThruster (_V(-1.913, 0, -0.61), _V( 0, 0,-1), RCS_TH, hProp, RCS_ISP);
	th_rcs[5] = CreateThruster (_V(-1.913, 0, -0.86), _V( 0, 0, 1), RCS_TH, hProp, RCS_ISP);
	th_rcs[6] = CreateThruster (_V(-1.913,-0.13, -0.73), _V( 0, 1, 0), RCS_TH, hProp, RCS_ISP);
	th_rcs[7] = CreateThruster (_V(-1.913, 0.13, -0.73), _V( 0,-1, 0), RCS_TH, hProp, RCS_ISP);
	
	th_group[0] = th_rcs[ 3];
	th_group[1] = th_rcs[ 7];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_PITCHUP);

	th_group[0] = th_rcs[ 2];
	th_group[1] = th_rcs[ 6];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_PITCHDOWN);
 
	th_group[0] = th_rcs[ 1];
	th_group[1] = th_rcs[ 4];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_YAWLEFT);

	th_group[0] = th_rcs[ 0];
	th_group[1] = th_rcs[ 5];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_YAWRIGHT);

	th_group[0] = th_rcs[ 3];
	th_group[1] = th_rcs[ 6];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_BANKRIGHT);

	th_group[0] = th_rcs[ 2];
	th_group[1] = th_rcs[ 7];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_BANKLEFT);

	th_group[0] = th_rcs[ 1];
	th_group[1] = th_rcs[ 5];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_FORWARD);

	th_group[0] = th_rcs[ 0];
	th_group[1] = th_rcs[ 4];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_BACK);

	th_group[0] = th_rcs[ 1];
	th_group[1] = th_rcs[ 5];
	CreateThrusterGroup (th_group, 2, THGROUP_MAIN);

	//////////////
	SURFHANDLE tex_rcs = oapiRegisterExhaustTexture ("exhaust_atrcs");
	double ex_lng = 2;
	double ex_wdt = 0.1;
	int i;
	for (i = 0; i < 8; i++)
	{
	AddExhaust (th_rcs[i], ex_lng, ex_wdt, (tex_rcs));
	}

	///////////////////////////////////////////////////////
	/////AirFoils


	hVA = CreateAirfoil3 (LIFT_VERTICAL, _V(0,0,-cap_ref), VLiftCoeff, 0, cap_size, cap_size, 1);
	hHA = CreateAirfoil3 (LIFT_HORIZONTAL, _V(0,0,-cap_ref), HLiftCoeff, 0, cap_size, cap_size, 1);

	

	////////////////////// boom pstreams

	SURFHANDLE exh = oapiRegisterExhaustTexture ("contrail3");

	PARTICLESTREAMSPEC big_pstream = {
		0, 2, 5, 10, 0.7, 5, 20, 4.0, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_PSQRT, 0, 8,
		PARTICLESTREAMSPEC::ATM_FLAT, 1, 1
		};

	PARTICLESTREAMSPEC small_pstream = {
		0, 0.25, 1, 25, 2, 2, 0, 4.0, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_PSQRT, 0, 1,
		PARTICLESTREAMSPEC::ATM_FLAT, 1, 1,
		exh};

	AddParticleStream(&small_pstream, _V(0,0,0), _V(0,0,-1), &pstream_lvl);
	AddParticleStream(&small_pstream, _V(0,0,0), _V(0,0,-1), &pstream_lvl);
	AddParticleStream(&small_pstream, _V(0,0,0), _V(0,1,0), &pstream_lvl);
	AddParticleStream(&small_pstream, _V(0,0,0), _V(0,-1,0), &pstream_lvl);
	AddParticleStream(&small_pstream, _V(0,0,0), _V(1,0,0), &pstream_lvl);
	AddParticleStream(&small_pstream, _V(0,0,0), _V(-1,0,0), &pstream_lvl);

	AddParticleStream(&big_pstream, _V(0,0,0), _V(0,0,1), &pstream_lvl);
	AddParticleStream(&big_pstream, _V(0,0,0), _V(0,0,-1), &pstream_lvl);
	AddParticleStream(&big_pstream, _V(0,0,0), _V(0,1,0), &pstream_lvl);
	AddParticleStream(&big_pstream, _V(0,0,0), _V(0,-1,0), &pstream_lvl);
	AddParticleStream(&big_pstream, _V(0,0,0), _V(1,0,0), &pstream_lvl);
	AddParticleStream(&big_pstream, _V(0,0,0), _V(-1,0,0), &pstream_lvl);


	////////// attachments
	hTD_att = CreateAttachment(false, _V(0,0,0), _V(0,0,1),  _V(0,1,0), "TD", false);
	hFH_att = CreateAttachment(true, _V(0,0,-1.632), _V(0,0,-1),  _V(0,1,0), "FH", false);

	if(col_status > 0)
	{
		DelAirfoil(hVA);
		DelAirfoil(hHA);
	}
}

// ==============================================================
// API callback interface
// ==============================================================

// --------------------------------------------------------------
// Vessel initialisation
// --------------------------------------------------------------
DLLCLBK VESSEL *ovcInit (OBJHANDLE hvessel, int flightmodel)
{
	return new titan_dragonfly_backshell (hvessel, flightmodel);
}




// --------------------------------------------------------------
// Vessel cleanup
// --------------------------------------------------------------
DLLCLBK void ovcExit (VESSEL *vessel)
{
	if (vessel) delete (titan_dragonfly_backshell*)vessel;
}
// ==============================================================
// Overloaded clbkPostCreation callback functions
// ==============================================================

void titan_dragonfly_backshell::clbkPostCreation ()
{
	CalcMass();
	if(skin_status == 1)ApplySkin();

	if (chute_status < 2) 
	{
		EditAirfoil (hVA,0x09, _V(0,0,-(cap_ref+(par_ref*(1-chute_proc)))), 0, 0, (cap_size+(par_size*(1-chute_proc))), 0);
		EditAirfoil (hHA,0x09, _V(0,0,-(cap_ref+(par_ref*(1-chute_proc)))), 0, 0, (cap_size+(par_size*(1-chute_proc))), 0);
		if (chute_status == 0) SetSize(15);
		SetPMI (_V(4,4,2));
	}

	if (drogue_status < 3 && drogue_status != 0)
	{
		EditAirfoil (hVA,0x09, _V(0,0,-(cap_ref+(drg_ref*(1-drogue_proc)))), 0, 0, (cap_size+(drg_size*(1-drogue_proc))), 0);
		EditAirfoil (hHA,0x09, _V(0,0,-(cap_ref+(drg_ref*(1-drogue_proc)))), 0, 0, (cap_size+(drg_size*(1-drogue_proc))), 0);
		SetAnimation (anim_drogue, drogue_proc);
		if (drogue_status == 1) SetSize(8);
	}

	if (cruise_status == 1) 
	{
		DelMesh(5);
		DeleteRCS();
	}
	if (lid_status == 1) DelMesh(2);
	if (htshld_status == 1) DelMesh(1);
	
	if (drogue_status == 0) DelMesh(4);
	else if (drogue_status < 3)SetMeshVisibilityMode (4, MESHVIS_EXTERNAL); //4

	if (chute_status < 2)SetMeshVisibilityMode (3, MESHVIS_EXTERNAL); //3

	if(cable_status > 0) SetAttachPoint();
	/////////////////////////////////////////
	MyID=ConnectToOrbiterSoundDLL(GetHandle());
	//ReplaceStockSound(MyID,"Sound\\Vessel\\ambient5.wav",	REPLACE_WHEELBRAKE);
	RequestLoadVesselWave(MyID,BOOM,"Sound\\Vessel\\sonicbangfar1.wav",BOTHVIEW_FADED_MEDIUM);
	ReplaceStockSound(MyID,"Sound\\Vessel\\ambient5.wav",	REPLACE_LAND_TOUCHDOWN);
	//ReplaceStockSound(MyID,"NULL",	REPLACE_LAND_TOUCHDOWN);
	SoundOptionOnOff(MyID,PLAYMAINTHRUST,FALSE);
	//SoundOptionOnOff(MyID,PLAYRETROTHRUST,FALSE);
	//SoundOptionOnOff(MyID,PLAYHOVERTHRUST,FALSE);
	SoundOptionOnOff(MyID,PLAYRADIOATC,FALSE);
	SoundOptionOnOff(MyID,PLAYGPWS,FALSE);
	SoundOptionOnOff(MyID,PLAYCOUNTDOWNWHENTAKEOFF,FALSE);
	//SoundOptionOnOff(MyID,PLAYLANDINGANDGROUNDSOUND,FALSE);
	SoundOptionOnOff(MyID,PLAYCABINAIRCONDITIONING,FALSE);
	SoundOptionOnOff(MyID,PLAYCABINRANDOMAMBIANCE,FALSE);
	SoundOptionOnOff(MyID,DISPLAYTIMER,FALSE);
	SoundOptionOnOff(MyID,PLAYDOCKLANDCLEARANCE,FALSE);
	SoundOptionOnOff(MyID,PLAYRADARBIP,FALSE);

	
}
void titan_dragonfly_backshell::clbkSaveState (FILEHANDLE scn)
{

	char cbuf[256];

	// default vessel parameters
	VESSEL3::clbkSaveState (scn);

	// custom parameters
	
	sprintf (cbuf, "%d %0.4f %d %0.4f %d %0.4f", drogue_status, drogue_proc, chute_status, chute_proc, col_status, col_proc);
	oapiWriteScenario_string (scn, "CHTS", cbuf);

	sprintf (cbuf, "%d %0.4f", cable_status, cable_proc);
	oapiWriteScenario_string (scn, "CABL", cbuf);

	sprintf (cbuf, "%d %d %d %d", cruise_status, lid_status, htshld_status, td_status);
	oapiWriteScenario_string (scn, "PLDS", cbuf);

	sprintf (cbuf, "%d %d", descent_status, attitude_status);
	oapiWriteScenario_string (scn, "EDL", cbuf);

	sprintf (cbuf, "%d", skin_status);
	oapiWriteScenario_string (scn, "CHAR", cbuf);
}
void titan_dragonfly_backshell::clbkLoadStateEx (FILEHANDLE scn, void *vs)
{
    char *line;

	while (oapiReadScenario_nextline (scn, line)) {
		if (!_strnicmp (line, "CHTS", 4))
		{
		sscanf (line+4, "%d%lf%d%lf%d%lf", &drogue_status, &drogue_proc, &chute_status, &chute_proc, &col_status, &col_proc);
		}
		else if (!_strnicmp (line, "CABL", 4))
		{
		sscanf (line+4, "%d%lf", &cable_status, &cable_proc);
		}
		else if (!_strnicmp (line, "PLDS", 4))
		{
		sscanf (line+4, "%d%d%d%d", &cruise_status, &lid_status, &htshld_status, &td_status);
		}
		else if (!_strnicmp (line, "EDL", 3))
		{
		sscanf (line+3, "%d%d", &descent_status, &attitude_status);
		}
		else if (!_strnicmp (line, "CHAR", 4))
		{
		sscanf (line+4, "%d", &skin_status);
		}
		else
		{
            ParseScenarioLineEx (line, vs);
			// unrecognised option - pass to Orbiter's generic parser
        }
   }
	SetAnimation (anim_chute, chute_proc);
	SetAnimation (anim_col, col_proc);
	SetAnimation (anim_drogue, drogue_proc);
	SetAnimation (anim_cable, cable_proc);
	
}

int titan_dragonfly_backshell::clbkConsumeBufferedKey(DWORD key, bool down, char *kstate)
{
  if (!down) return 0; // only process keydown events

  if (KEYMOD_SHIFT (kstate)) // +shift
	{
	  /*switch (key) 
	  {
		case OAPI_KEY_NUMPAD5: 
		if (kilrot_sw  == 0)kilrot_sw = 1;
		else kilrot_sw = 0;
		return 1;
	  }*/
	  return 0;
	} 
  else if (KEYMOD_CONTROL (kstate)) // +ctrl
	{
	return 0;
	}
  else // unmodified keys
	{ 
    switch (key) 
		{
		case OAPI_KEY_J:
		if (cruise_status  == 0)JettisonCruise();
		else if (htshld_status  == 0)JettisonHeatshield();
		else{
			DetachChild(hTD_att);
			td_status = 1;
			}
		return 1;

		case OAPI_KEY_P: // descent autopilot
		if (descent_status  == 0)descent_status = 1;
		else 
		{
			descent_status = 0;
			KillRCS();
		}
		return 1;

		case OAPI_KEY_K:
		if (cruise_status  == 1)
		{
		if(drogue_status == 3)
			{
			drogue_status = 2;
			SetMeshVisibilityMode(4, MESHVIS_EXTERNAL); 
			}
		else if(chute_status == 2)
			{
			chute_status = 1;
			SetMeshVisibilityMode(3, MESHVIS_EXTERNAL);
			drogue_status = 0;
			DelMesh(4);
			}
		else if(cable_status == 0)cable_status = 1;
		}
		return 1;


		}
	return 0;
	}
  
 return 0;
 }


bool titan_dragonfly_backshell::clbkDrawHUD (int mode, const HUDPAINTSPEC *hps, oapi::Sketchpad *skp)
{

	// draw the default HUD
	VESSEL3::clbkDrawHUD (mode, hps, skp);

	int s = hps->H;
	double d = (s*0.00130208);
	int sw = ((hps->W));
	int lw =  (int)(16*sw/1024);
	int lwoffset = sw - (18*lw);
	int hlw = (int) (lw/2);


	int roxl = 0;
	int royl = 0;

	double ds = s;
	double dsw = sw;
	double sc_ratio = ds/dsw;


	if (sc_ratio < 0.7284)
	{
	roxl = (lw*10);
	royl = (int) (-88*d);
	}



	//int wd = (int)(152*d);
	int wc = (int)(168*d);
	int w0 = (int)(184*d);
	int w1 = (int) (200*d);
	int w2 = (int)(216*d);
	int w3 = (int)(232*d);
	int w4 = (int)(248*d);
	/*int w5 = (int)(264*d);
	int w6 = (int)(280*d);
	int w7 = (int)(296*d);
	int w8 = (int)(312*d);
	int w9 = (int)(328*d);
	int w10 = (int)(344*d);
	int w11 = (int)(360*d);
	int w12 = (int)(376*d);
	int w13 = (int)(392*d);
	int w14 = (int)(408*d);
	int w15 = (int)(424*d);*/

	if(cruise_status == 0)
	{
	skp->SetTextColor(0x0066FF66);
	double m1 = GetEmptyMass();
	double m2 = GetPropellantMass(hProp);
	double mass_fraction = ((m1+m2)/(m1));
	double dv = (RCS_ISP*(log(mass_fraction)));
	char cbuf[256];
	sprintf(cbuf,"Available dV %0.1f m/s", dv);
	skp->Text ((10+roxl), (wc+royl), cbuf,strlen(cbuf));
	}

	
	{
	skp->SetTextColor(0x0066FF66);
	if(descent_status > 0)skp->SetTextColor(0x0066FFFF);
	char abuf[256];
	if(descent_status == 0)sprintf(abuf,"P = EDL Sequence:ON/OFF");
	else if(descent_status == 1)sprintf(abuf,"Next EDL Event: 7000km->EDL ATTITUDE");
	else if(descent_status == 2)sprintf(abuf,"Next EDL Event: 5500km->CRUISE SEP");
	else if(descent_status == 3)sprintf(abuf,"Next EDL Event: Mach 1.5->DROGUE");
	else if(descent_status == 4)sprintf(abuf,"Next EDL Event: 4km->CHUTE");
	else if(descent_status == 5)sprintf(abuf,"Next EDL Event: 3.8km->HEATSHIELD SEP");
	else if(descent_status == 6)sprintf(abuf,"Next EDL Event: 3.5km->LANDER SKIS");
	else if(descent_status == 7)sprintf(abuf,"Next EDL Event: 1.30km->LOWER LANDER");
	else if(descent_status == 8)sprintf(abuf,"Next EDL Event: 1.25km->LANDER ROTORS");
	else if(descent_status == 9)sprintf(abuf,"Next EDL Event: 1.20km->LANDER SEP");
	else if(descent_status == 10)sprintf(abuf,"Next EDL Event: TOUCHDOWN");
	else if(descent_status == 11)sprintf(abuf,"EDL FINISHED");
	skp->Text ((10+roxl), (w1+royl), abuf,strlen(abuf));
	}

	{
	skp->SetTextColor(0x0066FF66);
	//if(descent_status > 0)skp->SetTextColor(0x0066FFFF);
	char abuf[256];
	if(cruise_status == 0)
		{
		sprintf(abuf,"J = Manual Jettison Cruise_Stage");
		skp->Text ((10+roxl), (w3+royl), abuf,strlen(abuf));
		}
	else if(htshld_status == 0)
		{
		sprintf(abuf,"J = Manual Jettison Heatshield");
		skp->Text ((10+roxl), (w3+royl), abuf,strlen(abuf));
		}
	else if(td_status == 0)
		{
		sprintf(abuf,"J = Manual Jettison Lander");
		skp->Text ((10+roxl), (w3+royl), abuf,strlen(abuf));
		}
	}

	{
	skp->SetTextColor(0x0066FF66);
	//if(descent_status > 0)skp->SetTextColor(0x0066FFFF);
	char abuf[256];
	if(drogue_status == 3)
		{
		sprintf(abuf,"K = Manual Deploy Drogue");
		skp->Text ((10+roxl), (w4+royl), abuf,strlen(abuf));
		}
	else if(chute_status == 2)
		{
		sprintf(abuf,"K = Manual Deploy Parachute");
		skp->Text ((10+roxl), (w4+royl), abuf,strlen(abuf));
		}
	}

	

	return true;
}


void titan_dragonfly_backshell::clbkPreStep(double SimT, double SimDT, double MJD) {

	step_time = SimDT;

	double alti = GetAltitude(ALTMODE_MEANRAD);
	if(alti < 10000)alti = GetAltitude(ALTMODE_GROUND);
	//double aspd = 0;
	//aspd = GetAirspeed();

	if(GetAttachmentStatus(hFH_att) == NULL)
	{
	if (boom_status == 0)
		{
		if(pHndl == NULL)
			{
		// reentry pstream
			PARTICLESTREAMSPEC rps = {
			0, 2.5, 40, 0, 0.001, 0.3, 4, 0.2, PARTICLESTREAMSPEC::EMISSIVE,
			PARTICLESTREAMSPEC::LVL_PSQRT, 0, 1, PARTICLESTREAMSPEC::ATM_PLOG, 1e7, 1e16};
			pHndl = AddReentryStream (&rps);
			}
	////////////////////////// boom

		if (GetDynPressure() > 40000) 
			{
			boom_status = 1;
			}
		}
	else if (boom_status == 1) DoBoom();
	else if (boom_status == 2) return;

	///////////////////////////////// skin
		if(skin_status == 0)
		{
		if(GetDynPressure() > 4000) ApplySkin();
		}
	
	/////////////////////////////// descent
	
	if (descent_status == 1) 
	{
		if(cruise_status == 0)
		{
		if(alti < 7000000) GoToEDLAttitude();
		}
		else descent_status = 3;
	}
	else if (descent_status == 2)
	{
		if(alti < 5500000)
		{
		if(cruise_status == 0)
			{
			JettisonCruise();
			descent_status = 3;
			}
		}
	}
	else if (descent_status == 3)
	{
		if(drogue_status < 3)descent_status = 4;
		else if(alti < 150000 && GetMachNumber() < 1.5) //drogue
		{
		SetMeshVisibilityMode (4, MESHVIS_EXTERNAL);
		DelMesh(2);
		lid_status = 1;
		drogue_status = 2;
		descent_status = 4;
		}
	}
	else if (descent_status == 4)
	{
		if(chute_status < 2)descent_status = 5;
		else if(alti < 4000) //chute
		{
		SetMeshVisibilityMode (3, MESHVIS_EXTERNAL);
		DelMesh(4);
		drogue_status = 0;
		chute_status = 1;
		descent_status = 5;
		}
	}
	else if (descent_status == 5)
	{
		if(htshld_status == 1)descent_status = 6;
		else if(alti < 3800) //hs
		{
		JettisonHeatshield();
		descent_status = 6;
		}
	}
	else if (descent_status == 6)
	{
	if(alti < 3500) // lander legs
		{
		OBJHANDLE hTiDr = GetAttachmentStatus(hTD_att);
		VESSEL2* v =(VESSEL2*)oapiGetVesselInterface(hTiDr);
		v->SendBufferedKey(OAPI_KEY_G);
		descent_status = 7;
		}
	}
	else if (descent_status == 7)
	{
	if(alti < 1300) // lander pose
		{
		if(cable_status == 0)cable_status = 1;
		if(cable_status == 2)descent_status = 8;
		}
	}
	else if (descent_status == 8)
	{
	if(alti < 1250) // lander spinup
		{
		OBJHANDLE hTiDr = GetAttachmentStatus(hTD_att);
		VESSEL2* v =(VESSEL2*)oapiGetVesselInterface(hTiDr);
		//v->SetThrusterGroupLevel(THGROUP_MAIN, 0.2);
		v->SendBufferedKey(OAPI_KEY_N);
		descent_status = 9;
		}
	}
	else if (descent_status == 9)
	{
		if(alti < 1200) //lander
		{
		OBJHANDLE hTD = GetAttachmentStatus(hTD_att);
		DetachChild(hTD_att, 0.1);
		oapiSetFocusObject(hTD);
		descent_status = 10;
		td_status = 1;
		}
	}
	else if (descent_status == 10)
	{
	if(GroundContact() == true)
		{
		if (col_status == 0 && chute_status == 0) col_status = 1;
		DelAirfoil(hVA);
		DelAirfoil(hHA);
		landed_sw = 1;
		descent_status = 11;
		}

	}
	}

	if(GroundContact() == true) // not on autop
		{
		if (col_status == 0 && chute_status == 0) col_status = 1;
		DelAirfoil(hVA);
		DelAirfoil(hHA);
		landed_sw = 1;
		}

	//////////////// CHUTES /////////////////////////////////////
	
	if (chute_status == 2)
	{
		chute_proc = 1;
		SetAnimation (anim_chute, chute_proc);
	}
	else if (chute_status == 1) 
	{
		double da = SimDT * CHUTE_OPERATING_SPEED;
		chute_proc = chute_proc - da;
		if (chute_proc < 0) 
		{
		chute_proc = 0;
		chute_status = 0;
		SetPMI (_V(4,4,2));
		SetSize(15);
		}

		EditAirfoil (hVA,0x09, _V(0,0,-(cap_ref+(par_ref*(1-chute_proc)))), 0, 0, (cap_size+(par_size*(1-chute_proc))), 0);
		EditAirfoil (hHA,0x09, _V(0,0,-(cap_ref+(par_ref*(1-chute_proc)))), 0, 0, (cap_size+(par_size*(1-chute_proc))), 0);

		SetAnimation (anim_chute, chute_proc);
	}

	///// collapse chute
	if (col_status == 1)
	{
	double dc = SimDT * COLLAPSE_SPEED;
	col_proc = col_proc + dc;
	if (col_proc > 1)
		{
		col_proc = 1;
		col_status = 2;
		}
	SetAnimation (anim_col, col_proc);
	}
	else if (col_status == 2)
	{
	col_proc = 1;
	SetAnimation (anim_col, col_proc);
	}
	////////////////////////drogue ////////////////////////////

	if (drogue_status == 3)
	{
		drogue_proc = 1;
		SetAnimation (anim_drogue, drogue_proc);
	}
	else if (drogue_status == 2) 
	{
		double da = SimDT * CHUTE_OPERATING_SPEED;
		drogue_proc = drogue_proc - da;
		if (drogue_proc < 0)
		{
		drogue_proc = 0;
		drogue_status = 1;
		SetSize(8);
		}

		EditAirfoil (hVA,0x09, _V(0,0,-(cap_ref+(drg_ref*(1-drogue_proc)))), 0, 0, (cap_size+(drg_size*(1-drogue_proc))), 0);
		EditAirfoil (hHA,0x09, _V(0,0,-(cap_ref+(drg_ref*(1-drogue_proc)))), 0, 0, (cap_size+(drg_size*(1-drogue_proc))), 0);
		SetAnimation (anim_drogue, drogue_proc);
	}
	///////////////////////////////

	/////////////// cables ////////////////////

	if(cable_status == 0)cable_proc = 0;
	else if(cable_status == 1)
	{
		double da = SimDT * 0.5;
		cable_proc = cable_proc + da;
		if(cable_proc > 1)
		{
		cable_proc = 1;
		cable_status = 2;
		}
		SetAttachPoint();
		SetAnimation(anim_cable, cable_proc);
	}
	else if(cable_status == 2)
	{
		cable_proc = 1;
		SetAttachPoint();
		SetAnimation(anim_cable, cable_proc);
	}


}
void titan_dragonfly_backshell::clbkPostStep (double simt, double simdt, double mjd)
{
	if(GroundContact() == true) // not on autop
		{
		if (col_status == 0 && chute_status == 0) col_status = 1;
		DelAirfoil(hVA);
		DelAirfoil(hHA);
		landed_sw = 1;
		}

	if(landed_sw == 1)
	{
	VESSELSTATUS2 vs;
	memset (&vs, 0, sizeof(vs));
	vs.version = 2;
	vs.rbody = GetGravityRef();
	vs.status = 1;
	OBJHANDLE hVessel = GetHandle();
	double heading;
	oapiGetHeading(hVessel,&heading);
	double lng, lat, rad;
	GetEquPos(lng, lat, rad);
	vs.surf_lat = lat;
	vs.surf_lng = lng;
	vs.surf_hdg = heading;
	vs.arot.x = 10;
	DefSetStateEx (&vs);
	landed_sw = 2;
	}
}

void titan_dragonfly_backshell::CalcMass()
{
	double current_mass = BKSHL_MASS;
	if(htshld_status == 0)current_mass = current_mass + HTSHLD_MASS;
	if(cruise_status == 0)current_mass = current_mass + CRUISE_MASS;
	if(td_status == 0)current_mass = current_mass + TD_MASS;
	SetEmptyMass(current_mass);
}

////////// JETTISONS //////////////////////////////

void titan_dragonfly_backshell::JettisonHeatshield()
{
	VECTOR3 ofs = _V(0,0,0);
	VECTOR3 vofs = _V(0,0,0.2);
	char name[256];
	char classname[256];
	strcpy (classname, "titan_dragonfly/heatshield");
	strcpy (name, "Heatshield");

	
  VESSELSTATUS vs;
  GetStatus(vs);
  VECTOR3 rvel;
  Local2Rel (ofs, vs.rpos);
  GlobalRot (vofs, rvel);
  vs.rvel = vs.rvel + rvel;
  vs.status = 0;
  oapiCreateVessel (name, classname, vs);

  DelMesh(1);
  htshld_status = 1;
  CalcMass();

}

void titan_dragonfly_backshell::JettisonCruise()
{
	VECTOR3 ofs = _V(0,0,-1.2);
	VECTOR3 vofs = _V(0,0,-0.5);
	char name[256];
	char classname[256];
	strcpy (classname, "titan_dragonfly/cruise_stage");
	strcpy (name, "CruiseStage");

	
  VESSELSTATUS vs;
  GetStatus(vs);
  VECTOR3 rvel;
  Local2Rel (ofs, vs.rpos);
  GlobalRot (vofs, rvel);
  vs.rvel = vs.rvel + rvel;
  vs.status = 0;

  double prop_lvl = GetPropellantMass(hProp);
  PROPELLANT_HANDLE hPropC;
  OBJHANDLE hCruiseV = oapiCreateVessel (name, classname, vs);
  VESSEL2* v =(VESSEL2*)oapiGetVesselInterface(hCruiseV);
  hPropC = v->GetPropellantHandleByIndex(0);
  v->SetPropellantMass(hPropC, prop_lvl);
  

  DelMesh(5);
  cruise_status = 1;
  CalcMass();
  DeleteRCS();
}

void titan_dragonfly_backshell::DoBoom()
{
pstream_lvl = 1;
boom_proc = boom_proc+oapiGetSimStep();
if (boom_proc > 0.5)
	{
	pstream_lvl = 0;
	boom_proc = 0;
	boom_status = 2;
	DelExhaustStream(pHndl);
	PlayVesselWave(MyID, BOOM, 0, 255);
	OBJHANDLE hTD = GetAttachmentStatus(hTD_att);
	if(hTD) oapiDeleteVessel(hTD);
	SetMeshVisibilityMode (0, MESHVIS_NEVER);
	if(htshld_status == 0)SetMeshVisibilityMode (1, MESHVIS_NEVER);
	if(lid_status == 0)SetMeshVisibilityMode (2, MESHVIS_NEVER);
	SetMeshVisibilityMode (3, MESHVIS_NEVER);
	SetMeshVisibilityMode (4, MESHVIS_NEVER);
	if(cruise_status == 0)SetMeshVisibilityMode (5, MESHVIS_NEVER);
	}

}

void titan_dragonfly_backshell::GoToEDLAttitude()
{
	if(attitude_status == 3) // done prev
	{
	descent_status = 2;
	return;
	}

	VECTOR3 avel;
	GetAngularVel(avel);
	double mass = GetMass();

	if(attitude_status == 0)
	{
		if(length(avel) > (0.1*RAD))
		{
		if(GetNavmodeState(1) == false)
			{
			DeactivateNavmode(2);
			DeactivateNavmode(3);
			DeactivateNavmode(4);
			DeactivateNavmode(5);
			DeactivateNavmode(6);
			DeactivateNavmode(7);
			ActivateNavmode(1);
			}
		}
		else
		{
			DeactivateNavmode(1);
			attitude_status = 1;
		}

	}
	else if(attitude_status == 1)
	{
		VECTOR3 vel_rel, pos_glob;
		OBJHANDLE hTitan = oapiGetObjectByName("Titan");
		GetGlobalPos(pos_glob);
		oapiGetRelativeVel(GetHandle(), hTitan, &vel_rel);
		VECTOR3 tgtvec = pos_glob + vel_rel;
		VECTOR3 tgtvec_loc;
		Global2Local(tgtvec, tgtvec_loc);
		normalise(tgtvec_loc);
		double dAngle = acos(dotp(_V(0,0,1), tgtvec_loc));

		if(dAngle < (0.05*RAD) && length(avel) < (0.01*RAD))
		{
		attitude_status = 2;
		KillRCS();
		return;
		}
		
		
		double vrot_lim = (2*RAD);

		double max_acc_x = 2*20*0.73/(mass*1.4);
		double max_acc_y = 2*20*1.913/(mass*1.4);
		

		double max_drot_x =  max_acc_x*step_time;
		double max_drot_y =  max_acc_y*step_time;
		

		double angoa = atan2(-tgtvec_loc.y, tgtvec_loc.z);
		double slipage = atan2(tgtvec_loc.x, tgtvec_loc.z);

		if(angoa > PI)angoa = (PI*2)-angoa;
		else if(angoa < -PI)angoa = (PI*2)+angoa;

		if(slipage > PI)slipage = (PI*2)-slipage;
		else if(slipage < -PI)slipage = (PI*2)+slipage;

double reqxv = 0.9*(sqrt(abs(angoa*max_acc_x*2)));
if (reqxv > vrot_lim)reqxv = vrot_lim;
if (angoa > 0)reqxv = -reqxv;
double req_drotx = reqxv - avel.x;
double thx_lvl = req_drotx/max_drot_x;
if((angoa*DEG) < 1 && (angoa*DEG) > -1)
{
reqxv = 1*-angoa;
double acc_reqx = (reqxv-avel.x)/0.2;
thx_lvl = acc_reqx/max_acc_x;
}

double reqyv = 0.9*(sqrt(abs(slipage*max_acc_y*2)));
if (reqyv > vrot_lim)reqyv = vrot_lim;
if (slipage > 0)reqyv = -reqyv;
double req_droty = reqyv - avel.y;
double thy_lvl = req_droty/max_drot_y;
if((slipage*DEG) < 1 && (slipage*DEG) > -1)
{
reqyv = 1*-slipage;
double acc_reqy = (reqyv-avel.y)/0.2;
thy_lvl = acc_reqy/max_acc_y;
}

	//////////////////////////////////////////////
if(thx_lvl > 1)thx_lvl = 1;
else if(thx_lvl < -1)thx_lvl = -1;

if(thy_lvl > 1)thy_lvl = 1;
else if(thy_lvl < -1)thy_lvl = -1;


if(thx_lvl > 0)
{
SetThrusterGroupLevel(THGROUP_ATT_PITCHUP, thx_lvl);
SetThrusterGroupLevel(THGROUP_ATT_PITCHDOWN, 0);
}
else if(thx_lvl < 0)
{
SetThrusterGroupLevel(THGROUP_ATT_PITCHUP, 0);
SetThrusterGroupLevel(THGROUP_ATT_PITCHDOWN, -thx_lvl);
}
else
{
SetThrusterGroupLevel(THGROUP_ATT_PITCHUP, 0);
SetThrusterGroupLevel(THGROUP_ATT_PITCHDOWN, 0);
}


if(thy_lvl > 0)
{
SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT, thy_lvl);
SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT, 0);
}
else if(thy_lvl < 0)
{
SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT, 0);
SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT, -thy_lvl);
}
else
{
SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT, 0);
SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT, 0);
}

	

	}
	else if(attitude_status == 2)
	{
		double zvel_req = PI2/60;
		double drot_req = zvel_req - avel.z;
		double max_acc_z = 2*20*1.913/(mass*1.5);
		double max_drot_z =  max_acc_z*step_time;

		double th_lvl = drot_req/max_drot_z;
		if(th_lvl > 1)th_lvl = 1;
		else if(th_lvl < -1)th_lvl = -1;

		if(abs(th_lvl) < 0.05)
		{
			KillRCS();
			descent_status = 2;
			attitude_status = 3;
			return;
		}

		if(th_lvl > 0)
		{
		SetThrusterGroupLevel(THGROUP_ATT_BANKRIGHT, th_lvl);
		SetThrusterGroupLevel(THGROUP_ATT_BANKLEFT, 0);
		}
		else if(th_lvl < 0)
		{
		SetThrusterGroupLevel(THGROUP_ATT_BANKRIGHT, 0);
		SetThrusterGroupLevel(THGROUP_ATT_BANKLEFT, -th_lvl);
		}
		else
		{
		SetThrusterGroupLevel(THGROUP_ATT_BANKLEFT, 0);
		SetThrusterGroupLevel(THGROUP_ATT_BANKRIGHT, 0);
		}

	}

	
}

void titan_dragonfly_backshell::KillRCS ()
{
SetThrusterGroupLevel(THGROUP_ATT_PITCHUP, 0);
SetThrusterGroupLevel(THGROUP_ATT_PITCHDOWN, 0);
SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT, 0);
SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT, 0);
SetThrusterGroupLevel(THGROUP_ATT_BANKRIGHT, 0);
SetThrusterGroupLevel(THGROUP_ATT_BANKLEFT, 0);
}

void titan_dragonfly_backshell::DeleteRCS()
{
	DelThrusterGroup(THGROUP_ATT_PITCHUP, true);
	DelThrusterGroup(THGROUP_ATT_PITCHDOWN, true);
	DelThrusterGroup(THGROUP_ATT_YAWLEFT, true);
	DelThrusterGroup(THGROUP_ATT_YAWRIGHT, true);
	DelThrusterGroup(THGROUP_ATT_BANKLEFT, true);
	DelThrusterGroup(THGROUP_ATT_BANKRIGHT, true);
	DelThrusterGroup(THGROUP_ATT_FORWARD, true);
	DelThrusterGroup(THGROUP_ATT_BACK, true);
	DelThrusterGroup(THGROUP_MAIN, true);
	DelPropellantResource(hProp);
}

void titan_dragonfly_backshell::ApplySkin ()
{	
	 bs1 = oapiGetTextureHandle (hMbackshell, 2);
	oapiBlt(bs1, bs2, 0, 0, 0, 0, 512, 512);

	bs3 = oapiGetTextureHandle (hMHeatshield, 1);
	oapiBlt(bs3, bs4, 0, 0, 0, 0, 512, 512);

	skin_status = 1;
}


void titan_dragonfly_backshell::SetAttachPoint()
{
	VECTOR3 newpos = _V(0,0,(cable_proc * 0.5));
	SetAttachmentParams(hTD_att, newpos, _V(0,0,1), _V(0,1,0));
}