// ==============================================================
//                 ORBITER MODULE: titan_dragonfly_cruise
//                  
//
// titan_dragonfly_cruise.cpp
// Control module for titan_dragonfly_cruise vessel class
//
// 
// ==============================================================

#define STRICT
#define ORBITER_MODULE

#include "orbitersdk.h"
#include "OrbiterSoundSDK40.h"
//#include <cstdio>
//#include <cmath>
//#include <string>
#include <stdio.h>
#include <math.h>

// ==============================================================
// Some vessel parameters
// ==============================================================


const double CRUISE_MASS = 320;
const double FUEL_MASS = 400;
const double RCS_ISP = 3000;
const double RCS_TH = 20;

#define BOOM	1

double stiffness = 3e4;// 6e4
double damping = 1e4; // 5e4

static TOUCHDOWNVTX tdvtx_geardown[7] = {
	{_V( 0,2, 0.1),     stiffness, damping, 0.9, 0.2},
	{_V(-1,-1.732, 0.1), stiffness, damping, 0.9, 0.2},
	{_V( 1,-1.732, 0.1), stiffness, damping, 0.9, 0.2},
	{_V( 0,-2, 0.1),     stiffness, damping, 0.9},
	{_V(-1, 1.732, 0.1), stiffness, damping, 0.9},
	{_V( 1, 1.732, 0.1), stiffness, damping, 0.9},
	{_V(0,0, -2),        stiffness, damping, 0.9}
};


// Calculate lift coefficient [Cl] as a function of aoa (angle of attack) over -Pi ... Pi
// Implemented here as a piecewise linear function
void VLiftCoeff (VESSEL *v, double aoa, double M, double Re, void *context, double *cl, double *cm, double *cd)
{
	int i;
	const int nabsc = 9;
	static const double AOA[nabsc] = {-180*RAD,-60*RAD,-30*RAD, -2*RAD, 10*RAD,15*RAD,50*RAD,90*RAD,180*RAD};
	static const double CL[nabsc]  = {0,0,0,0,0,0,0,0,0};
	static const double CM[nabsc]  = {0,0,0,0,0,0,0,0,0};
	for (i = 0; i < nabsc-1 && AOA[i+1] < aoa; i++);
	double f = (aoa-AOA[i]) / (AOA[i+1]-AOA[i]);
	*cl = CL[i] + (CL[i+1]-CL[i]) * f;  // aoa-dependent lift coefficient
	*cm = 0.0;
	*cd = 0.95 + oapiGetInducedDrag (*cl, 1.5, 0.7) + oapiGetWaveDrag (M, 0.75, 1.0, 1.1, 0.04);
}

void HLiftCoeff (VESSEL *v, double beta, double M, double Re, void *context, double *cl, double *cm, double *cd)
{
	int i;
	const int nabsc = 8;
	static const double BETA[nabsc] = {-180*RAD,-135*RAD,-90*RAD,-45*RAD,45*RAD,90*RAD,135*RAD,180*RAD};
	static const double CL[nabsc]   = {0,0,0,0,0,0,0,0};
	for (i = 0; i < nabsc-1 && BETA[i+1] < beta; i++);
	*cl = CL[i] + (CL[i+1]-CL[i]) * (beta-BETA[i]) / (BETA[i+1]-BETA[i]);
	*cm = 0.0;
	*cd = 0.95 + oapiGetInducedDrag (*cl, 1.5, 0.7) + oapiGetWaveDrag (M, 0.75, 1.0, 1.1, 0.04);
}


class titan_dragonfly_cruise: public VESSEL3 {
public:
	
	titan_dragonfly_cruise (OBJHANDLE hVessel, int flightmodel);
	~titan_dragonfly_cruise ();

	void clbkSetClassCaps (FILEHANDLE cfg);
	void clbkPreStep(double SimT, double SimDT, double MJD);
//	void clbkPostStep (double simt, double simdt, double mjd);
	void clbkPostCreation();
	void clbkSaveState (FILEHANDLE scn);
	void clbkLoadStateEx (FILEHANDLE scn, void *status);
//	int clbkConsumeBufferedKey(DWORD key, bool down, char *kstate);
	void DoBoom(void);
	int MyID;
	

private:
	double boom_proc, pstream_lvl;
	int boom_status;
	AIRFOILHANDLE hHA, hVA;

	PSTREAM_HANDLE pHndl;
	THRUSTER_HANDLE th_rcs[8], th_group[2];
	PROPELLANT_HANDLE hProp;
	
};

titan_dragonfly_cruise::titan_dragonfly_cruise (OBJHANDLE hVessel, int flightmodel)
: VESSEL3 (hVessel, flightmodel)
{
}

titan_dragonfly_cruise::~titan_dragonfly_cruise ()
{
	
}


// ==============================================================
// titan_dragonfly_cruise class interface
// ==============================================================

// ==============================================================
// Overloaded callback functions
// ==============================================================

// --------------------------------------------------------------
// Set the capabilities of the vessel class
// --------------------------------------------------------------
void titan_dragonfly_cruise::clbkSetClassCaps (FILEHANDLE cfg)
{
	boom_proc = 0;
	pstream_lvl = 0;
	boom_status = 0;

	// physical specs
	SetCW (4, 4, 2,2);
	SetRotDrag (_V(2,2,0.1));
	SetPMI (_V(1.4,1.4,1.5));
	SetCrossSections (_V(2,2,10));
	SetSize (2);
	SetEmptyMass(CRUISE_MASS);
	SetCameraOffset (_V(0,0,0));
	SetTouchdownPoints ( tdvtx_geardown, 7);

	SetMeshVisibilityMode (AddMesh ("titan_dragonfly/cruise"), MESHVIS_EXTERNAL); 
	

	//////////////////////////////////////////

	hProp = CreatePropellantResource (FUEL_MASS, FUEL_MASS);

	
	// ***************** thruster definitions *******************

	th_rcs[0] = CreateThruster (_V( 1.913, 0, 0.59), _V( 0, 0,-1), RCS_TH, hProp, RCS_ISP);
	th_rcs[1] = CreateThruster (_V( 1.913, 0,  0.34), _V( 0, 0, 1), RCS_TH, hProp, RCS_ISP);
	th_rcs[2] = CreateThruster (_V( 1.913,-0.13,  0.47), _V( 0, 1, 0), RCS_TH, hProp, RCS_ISP);
	th_rcs[3] = CreateThruster (_V( 1.913, 0.13,  0.47), _V( 0,-1, 0), RCS_TH, hProp, RCS_ISP);

	th_rcs[4] = CreateThruster (_V(-1.913, 0,  0.59), _V( 0, 0,-1), RCS_TH, hProp, RCS_ISP);
	th_rcs[5] = CreateThruster (_V(-1.913, 0,  0.34), _V( 0, 0, 1), RCS_TH, hProp, RCS_ISP);
	th_rcs[6] = CreateThruster (_V(-1.913,-0.13,  0.47), _V( 0, 1, 0), RCS_TH, hProp, RCS_ISP);
	th_rcs[7] = CreateThruster (_V(-1.913, 0.13,  0.47), _V( 0,-1, 0), RCS_TH, hProp, RCS_ISP);
	
	th_group[0] = th_rcs[ 2];
	th_group[1] = th_rcs[ 6];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_PITCHUP);

	th_group[0] = th_rcs[ 3];
	th_group[1] = th_rcs[ 7];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_PITCHDOWN);
 
	th_group[0] = th_rcs[ 1];
	th_group[1] = th_rcs[ 4];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_YAWLEFT);

	th_group[0] = th_rcs[ 0];
	th_group[1] = th_rcs[ 5];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_YAWRIGHT);

	th_group[0] = th_rcs[ 3];
	th_group[1] = th_rcs[ 6];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_BANKRIGHT);

	th_group[0] = th_rcs[ 2];
	th_group[1] = th_rcs[ 7];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_BANKLEFT);

	th_group[0] = th_rcs[ 1];
	th_group[1] = th_rcs[ 5];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_FORWARD);

	th_group[0] = th_rcs[ 0];
	th_group[1] = th_rcs[ 4];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_BACK);

	th_group[0] = th_rcs[ 1];
	th_group[1] = th_rcs[ 5];
	CreateThrusterGroup (th_group, 2, THGROUP_MAIN);

	//////////////
	SURFHANDLE tex_rcs = oapiRegisterExhaustTexture ("exhaust_atrcs");
	double ex_lng = 2;
	double ex_wdt = 0.1;
	int i;
	for (i = 0; i < 8; i++)
	{
	AddExhaust (th_rcs[i], ex_lng, ex_wdt, (tex_rcs));
	}

	///////////////////////////////////////////////////////
	/////AirFoils


	hVA = CreateAirfoil3 (LIFT_VERTICAL, _V(0,0,0), VLiftCoeff, 0, 5, 5, 1);
	hHA = CreateAirfoil3 (LIFT_HORIZONTAL, _V(0,0,0), HLiftCoeff, 0, 5, 5, 1);

	// reentry pstream
	PARTICLESTREAMSPEC rps = {
		0, 2.5, 40, 0, 0.001, 0.3, 4, 0.2, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_PSQRT, 0, 1, PARTICLESTREAMSPEC::ATM_PLOG, 1e7, 1e16
	};
	pHndl = AddReentryStream (&rps);

	////////////////////// boom pstreams

	SURFHANDLE exh = oapiRegisterExhaustTexture ("contrail3");

	PARTICLESTREAMSPEC big_pstream = {
		0, 2, 5, 10, 0.7, 5, 20, 4.0, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_PSQRT, 0, 8,
		PARTICLESTREAMSPEC::ATM_FLAT, 1, 1
		};

	PARTICLESTREAMSPEC small_pstream = {
		0, 0.25, 1, 25, 2, 2, 0, 4.0, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_PSQRT, 0, 1,
		PARTICLESTREAMSPEC::ATM_FLAT, 1, 1,
		exh};

	AddParticleStream(&small_pstream, _V(0,0,0), _V(0,0,-1), &pstream_lvl);
	AddParticleStream(&small_pstream, _V(0,0,0), _V(0,0,-1), &pstream_lvl);
	AddParticleStream(&small_pstream, _V(0,0,0), _V(0,1,0), &pstream_lvl);
	AddParticleStream(&small_pstream, _V(0,0,0), _V(0,-1,0), &pstream_lvl);
	AddParticleStream(&small_pstream, _V(0,0,0), _V(1,0,0), &pstream_lvl);
	AddParticleStream(&small_pstream, _V(0,0,0), _V(-1,0,0), &pstream_lvl);

	AddParticleStream(&big_pstream, _V(0,0,0), _V(0,0,1), &pstream_lvl);
	AddParticleStream(&big_pstream, _V(0,0,0), _V(0,0,-1), &pstream_lvl);
	AddParticleStream(&big_pstream, _V(0,0,0), _V(0,1,0), &pstream_lvl);
	AddParticleStream(&big_pstream, _V(0,0,0), _V(0,-1,0), &pstream_lvl);
	AddParticleStream(&big_pstream, _V(0,0,0), _V(1,0,0), &pstream_lvl);
	AddParticleStream(&big_pstream, _V(0,0,0), _V(-1,0,0), &pstream_lvl);

}

// ==============================================================
// API callback interface
// ==============================================================

// --------------------------------------------------------------
// Vessel initialisation
// --------------------------------------------------------------
DLLCLBK VESSEL *ovcInit (OBJHANDLE hvessel, int flightmodel)
{
	return new titan_dragonfly_cruise (hvessel, flightmodel);
}




// --------------------------------------------------------------
// Vessel cleanup
// --------------------------------------------------------------
DLLCLBK void ovcExit (VESSEL *vessel)
{
	if (vessel) delete (titan_dragonfly_cruise*)vessel;
}
// ==============================================================
// Overloaded clbkPostCreation callback functions
// ==============================================================

void titan_dragonfly_cruise::clbkPostCreation ()
{
	
	/////////////////////////////////////////
	MyID=ConnectToOrbiterSoundDLL(GetHandle());
	//ReplaceStockSound(MyID,"Sound\\Vessel\\ambient5.wav",	REPLACE_WHEELBRAKE);
	RequestLoadVesselWave(MyID,BOOM,"Sound\\Vessel\\sonicbangfar1.wav",BOTHVIEW_FADED_MEDIUM);
	ReplaceStockSound(MyID,"Sound\\Vessel\\ambient5.wav",	REPLACE_LAND_TOUCHDOWN);
	//ReplaceStockSound(MyID,"NULL",	REPLACE_LAND_TOUCHDOWN);
	//SoundOptionOnOff(MyID,PLAYMAINTHRUST,FALSE);
	//SoundOptionOnOff(MyID,PLAYRETROTHRUST,FALSE);
	//SoundOptionOnOff(MyID,PLAYHOVERTHRUST,FALSE);
	SoundOptionOnOff(MyID,PLAYRADIOATC,FALSE);
	SoundOptionOnOff(MyID,PLAYGPWS,FALSE);
	SoundOptionOnOff(MyID,PLAYCOUNTDOWNWHENTAKEOFF,FALSE);
	//SoundOptionOnOff(MyID,PLAYLANDINGANDGROUNDSOUND,FALSE);
	SoundOptionOnOff(MyID,PLAYCABINAIRCONDITIONING,FALSE);
	SoundOptionOnOff(MyID,PLAYCABINRANDOMAMBIANCE,FALSE);
	SoundOptionOnOff(MyID,DISPLAYTIMER,FALSE);
	SoundOptionOnOff(MyID,PLAYDOCKLANDCLEARANCE,FALSE);
	SoundOptionOnOff(MyID,PLAYRADARBIP,FALSE);

	
}
void titan_dragonfly_cruise::clbkSaveState (FILEHANDLE scn)
{

	//char cbuf[256];

	// default vessel parameters
	VESSEL3::clbkSaveState (scn);

	// custom parameters
	


}
void titan_dragonfly_cruise::clbkLoadStateEx (FILEHANDLE scn, void *vs)
{
    char *line;

	while (oapiReadScenario_nextline (scn, line)) {
		
		{
            ParseScenarioLineEx (line, vs);
			// unrecognised option - pass to Orbiter's generic parser
        }
   }
	
	
}

//int titan_dragonfly_cruise::clbkConsumeBufferedKey(DWORD key, bool down, char *kstate)
//{
//  if (!down) return 0; // only process keydown events
//
//  if (KEYMOD_SHIFT (kstate)) // +shift
//	{
//	  /*switch (key) 
//	  {
//		case OAPI_KEY_NUMPAD5: 
//		if (kilrot_sw  == 0)kilrot_sw = 1;
//		else kilrot_sw = 0;
//		return 1;
//	  }*/
//	  return 0;
//	} 
//  else if (KEYMOD_CONTROL (kstate)) // +ctrl
//	{
//	return 0;
//	}
//  else // unmodified keys
//	{ 
//    switch (key) 
//		{
//		
//
//		}
//	return 0;
//	}
//  
// return 0;
// }


void titan_dragonfly_cruise::clbkPreStep(double SimT, double SimDT, double MJD) {

	
	////////////////////////// boom
	if (boom_status == 0)
	{
	if (GetDynPressure() > 1000) boom_status = 1;
	}
	else if (boom_status == 1) DoBoom();

	///////////////////////////////


}
//void titan_dragonfly_cruise::clbkPostStep (double simt, double simdt, double mjd)
//{
	////////////// report
	/*double alti = GetAltitude();

	if(report_sw == 0 && alti < 125000)
	{
	VECTOR3 haspdv = _V(0,0,0);
	GetHorizonAirspeedVector(haspdv);
	double xz = length(_V(haspdv.x,0,haspdv.z));
	double fpa = atan2(haspdv.y,xz);
	fpa = fpa*DEG;
	
	double lngi = 0;
	double lati = 0;
	double radi = 0;
	oapiGetEquPos(GetHandle(),&lngi,&lati,&radi);

	
	double tgt_lng = -6.1*RAD;
	double tgt_lat = -1.9*RAD;
	double antang = oapiOrthodome(lngi, lati, tgt_lng, tgt_lat);
	antang = antang*DEG;
	report_sw = 1;
	sprintf(oapiDebugString(), "fpa %0.2f antang %0.2f alti %0.2f",fpa, antang, alti*0.001);
	}*/
//}


void titan_dragonfly_cruise::DoBoom()
{
pstream_lvl = 1;
boom_proc = boom_proc+oapiGetSimStep();
if (boom_proc > 0.5)
	{
	
	pstream_lvl = 0;
	boom_proc = 0;
	boom_status = 2;
	DelExhaustStream(pHndl);
	PlayVesselWave(MyID, BOOM, 0, 255);
	oapiDeleteVessel(GetHandle());
	}

}
