// ==============================================================
//                 ORBITER MODULE: titan_dragonfly_heatshield
//                  
//
// titan_dragonfly_heatshield.cpp
// Control module for titan_dragonfly_heatshield vessel class
//
// 
// ==============================================================

#define STRICT
#define ORBITER_MODULE

#include "orbitersdk.h"
#include "OrbiterSoundSDK40.h"
//#include <cstdio>
//#include <cmath>
//#include <string>
#include <stdio.h>
#include <math.h>

// ==============================================================
// Some vessel parameters
// ==============================================================


const double HTSHLD_MASS = 220;


double stiffness = 3e4;// 6e4
double damping = 1e4; // 5e4

static TOUCHDOWNVTX tdvtx_geardown[7] = {
	{_V( 0,2, 0.7),     stiffness, damping, 0.9, 0.2},
	{_V(-1,-1.732, 0.7), stiffness, damping, 0.9, 0.2},
	{_V( 1,-1.732, 0.7), stiffness, damping, 0.9, 0.2},
	{_V( 0,-2, 0.7),     stiffness, damping, 0.9},
	{_V(-1, 1.732, 0.7), stiffness, damping, 0.9},
	{_V( 1, 1.732, 0.7), stiffness, damping, 0.9},
	{_V(0,0, -0.2),        stiffness, damping, 0.9}
};


// Calculate lift coefficient [Cl] as a function of aoa (angle of attack) over -Pi ... Pi
// Implemented here as a piecewise linear function
void VLiftCoeff (VESSEL *v, double aoa, double M, double Re, void *context, double *cl, double *cm, double *cd)
{
	int i;
	const int nabsc = 9;
	static const double AOA[nabsc] = {-180*RAD,-60*RAD,-30*RAD, -2*RAD, 10*RAD,15*RAD,50*RAD,90*RAD,180*RAD};
	static const double CL[nabsc]  = {0,0,0,0,0,0,0,0,0};
	static const double CM[nabsc]  = {0,0,0,0,0,0,0,0,0};
	for (i = 0; i < nabsc-1 && AOA[i+1] < aoa; i++);
	double f = (aoa-AOA[i]) / (AOA[i+1]-AOA[i]);
	*cl = CL[i] + (CL[i+1]-CL[i]) * f;  // aoa-dependent lift coefficient
	*cm = 0.0;
	*cd = 0.95 + oapiGetInducedDrag (*cl, 1.5, 0.7) + oapiGetWaveDrag (M, 0.75, 1.0, 1.1, 0.04);
}

void HLiftCoeff (VESSEL *v, double beta, double M, double Re, void *context, double *cl, double *cm, double *cd)
{
	int i;
	const int nabsc = 8;
	static const double BETA[nabsc] = {-180*RAD,-135*RAD,-90*RAD,-45*RAD,45*RAD,90*RAD,135*RAD,180*RAD};
	static const double CL[nabsc]   = {0,0,0,0,0,0,0,0};
	for (i = 0; i < nabsc-1 && BETA[i+1] < beta; i++);
	*cl = CL[i] + (CL[i+1]-CL[i]) * (beta-BETA[i]) / (BETA[i+1]-BETA[i]);
	*cm = 0.0;
	*cd = 0.95 + oapiGetInducedDrag (*cl, 1.5, 0.7) + oapiGetWaveDrag (M, 0.75, 1.0, 1.1, 0.04);
}


class titan_dragonfly_heatshield: public VESSEL3 {
public:
	
	titan_dragonfly_heatshield (OBJHANDLE hVessel, int flightmodel);
	~titan_dragonfly_heatshield ();

	void clbkSetClassCaps (FILEHANDLE cfg);
	void clbkPreStep(double SimT, double SimDT, double MJD);
	void clbkPostStep (double simt, double simdt, double mjd);
//	void clbkPostCreation();
	void clbkSaveState (FILEHANDLE scn);
	void clbkLoadStateEx (FILEHANDLE scn, void *status);
//	int clbkConsumeBufferedKey(DWORD key, bool down, char *kstate);
///	void DoBoom(void);
//	int MyID;
	

private:
//	double boom_proc, pstream_lvl;
	int landed_sw;
	AIRFOILHANDLE hHA, hVA;

//	PSTREAM_HANDLE pHndl;
//	THRUSTER_HANDLE th_rcs[8], th_group[2];
//	PROPELLANT_HANDLE hProp;
	
};

titan_dragonfly_heatshield::titan_dragonfly_heatshield (OBJHANDLE hVessel, int flightmodel)
: VESSEL3 (hVessel, flightmodel)
{
}

titan_dragonfly_heatshield::~titan_dragonfly_heatshield ()
{
	
}


// ==============================================================
// titan_dragonfly_heatshield class interface
// ==============================================================

// ==============================================================
// Overloaded callback functions
// ==============================================================

// --------------------------------------------------------------
// Set the capabilities of the vessel class
// --------------------------------------------------------------
void titan_dragonfly_heatshield::clbkSetClassCaps (FILEHANDLE cfg)
{

	landed_sw = 0;
	// physical specs
	SetCW (4, 4, 2,2);
	SetRotDrag (_V(1,1,1));
	SetPMI (_V(4,4,4));
	SetCrossSections (_V(2,2,10));
	SetSize (2);
	SetEmptyMass(HTSHLD_MASS);
	SetCameraOffset (_V(0,0,0));
	SetTouchdownPoints (tdvtx_geardown, 7);

	SetMeshVisibilityMode (AddMesh ("titan_dragonfly/heatshield"), MESHVIS_EXTERNAL); 
	

	//////////////////////////////////////////

	
	///////////////////////////////////////////////////////
	/////AirFoils


	hVA = CreateAirfoil3 (LIFT_VERTICAL, _V(0,0,0), VLiftCoeff, 0, 0.5, 0.5, 1);
	hHA = CreateAirfoil3 (LIFT_HORIZONTAL, _V(0,0,0), HLiftCoeff, 0, 0.5, 0.5, 1);


}

// ==============================================================
// API callback interface
// ==============================================================

// --------------------------------------------------------------
// Vessel initialisation
// --------------------------------------------------------------
DLLCLBK VESSEL *ovcInit (OBJHANDLE hvessel, int flightmodel)
{
	return new titan_dragonfly_heatshield (hvessel, flightmodel);
}




// --------------------------------------------------------------
// Vessel cleanup
// --------------------------------------------------------------
DLLCLBK void ovcExit (VESSEL *vessel)
{
	if (vessel) delete (titan_dragonfly_heatshield*)vessel;
}
// ==============================================================
// Overloaded clbkPostCreation callback functions
// ==============================================================

//void titan_dragonfly_heatshield::clbkPostCreation ()
//{
//	
//	/////////////////////////////////////////
//	MyID=ConnectToOrbiterSoundDLL(GetHandle());
//	//ReplaceStockSound(MyID,"Sound\\Vessel\\ambient5.wav",	REPLACE_WHEELBRAKE);
//	RequestLoadVesselWave(MyID,BOOM,"Sound\\Vessel\\sonicbangfar1.wav",BOTHVIEW_FADED_MEDIUM);
//	ReplaceStockSound(MyID,"Sound\\Vessel\\ambient5.wav",	REPLACE_LAND_TOUCHDOWN);
//	//ReplaceStockSound(MyID,"NULL",	REPLACE_LAND_TOUCHDOWN);
//	//SoundOptionOnOff(MyID,PLAYMAINTHRUST,FALSE);
//	//SoundOptionOnOff(MyID,PLAYRETROTHRUST,FALSE);
//	//SoundOptionOnOff(MyID,PLAYHOVERTHRUST,FALSE);
//	SoundOptionOnOff(MyID,PLAYRADIOATC,FALSE);
//	SoundOptionOnOff(MyID,PLAYGPWS,FALSE);
//	SoundOptionOnOff(MyID,PLAYCOUNTDOWNWHENTAKEOFF,FALSE);
//	//SoundOptionOnOff(MyID,PLAYLANDINGANDGROUNDSOUND,FALSE);
//	SoundOptionOnOff(MyID,PLAYCABINAIRCONDITIONING,FALSE);
//	SoundOptionOnOff(MyID,PLAYCABINRANDOMAMBIANCE,FALSE);
//	SoundOptionOnOff(MyID,DISPLAYTIMER,FALSE);
//	SoundOptionOnOff(MyID,PLAYDOCKLANDCLEARANCE,FALSE);
//	SoundOptionOnOff(MyID,PLAYRADARBIP,FALSE);
//
//	
//}
void titan_dragonfly_heatshield::clbkSaveState (FILEHANDLE scn)
{

	//char cbuf[256];

	// default vessel parameters
	VESSEL3::clbkSaveState (scn);

	// custom parameters
	


}
void titan_dragonfly_heatshield::clbkLoadStateEx (FILEHANDLE scn, void *vs)
{
    char *line;

	while (oapiReadScenario_nextline (scn, line)) {
		
		{
            ParseScenarioLineEx (line, vs);
			// unrecognised option - pass to Orbiter's generic parser
        }
   }
	
	
}

//int titan_dragonfly_heatshield::clbkConsumeBufferedKey(DWORD key, bool down, char *kstate)
//{
//  if (!down) return 0; // only process keydown events
//
//  if (KEYMOD_SHIFT (kstate)) // +shift
//	{
//	  /*switch (key) 
//	  {
//		case OAPI_KEY_NUMPAD5: 
//		if (kilrot_sw  == 0)kilrot_sw = 1;
//		else kilrot_sw = 0;
//		return 1;
//	  }*/
//	  return 0;
//	} 
//  else if (KEYMOD_CONTROL (kstate)) // +ctrl
//	{
//	return 0;
//	}
//  else // unmodified keys
//	{ 
//    switch (key) 
//		{
//		
//
//		}
//	return 0;
//	}
//  
// return 0;
// }


void titan_dragonfly_heatshield::clbkPreStep(double SimT, double SimDT, double MJD)
{
//if(GroundContact() == true && landed_sw == 0)landed_sw = 1;
if(GetAltitude(ALTMODE_GROUND) < 3 && landed_sw == 0)landed_sw = 1;


}


void titan_dragonfly_heatshield::clbkPostStep (double simt, double simdt, double mjd)
{
//	if(GroundContact() == true && landed_sw == 0)landed_sw = 1;
	if(GetAltitude(ALTMODE_GROUND) < 3 && landed_sw == 0)landed_sw = 1;

	if(landed_sw == 1)
	{
	VESSELSTATUS2 vs;
	memset (&vs, 0, sizeof(vs));
	vs.version = 2;
	vs.rbody = GetGravityRef();
	vs.status = 1;
	OBJHANDLE hVessel = GetHandle();
	double heading;
	oapiGetHeading(hVessel,&heading);
	double lng, lat, rad;
	GetEquPos(lng, lat, rad);
	vs.surf_lat = lat;
	vs.surf_lng = lng;
	vs.surf_hdg = heading;
	vs.arot.x = 10;
	DefSetStateEx (&vs);
	landed_sw = 2;
	}
}


//void titan_dragonfly_heatshield::DoBoom()
//{
//pstream_lvl = 1;
//boom_proc = boom_proc+oapiGetSimStep();
//if (boom_proc > 0.5)
//	{
//	
//	pstream_lvl = 0;
//	boom_proc = 0;
//	boom_status = 2;
//	DelExhaustStream(pHndl);
//	PlayVesselWave(MyID, BOOM, 0, 255);
//	}
//
//}
